#include <stdio.h>
#include <time.h>
#include <sys/poll.h>

/*  uses JDG's arm_v3

notes:

  random changed to rand to allow linking
  N.B. check rand as may not produce 32-bit no.s


needs:

  1. long multiply  - draughted BUT NOT TESTED
  2. btb updating for thumb
  3. more thumb testing

lsl routine added but not fully tested (28/10/98)
Long multiplication written (27/1/00)
Architecture V5 added (2/2/00)
Minor change to dis_transfer_offset (30/8/00)

To do:
   check long muls
         Flag checking (immediates ?!)
      Validation
   interrupt enable behaviour on exceptions (etc.)
*/


/* NB "int" is assumed to be at least 32-bit */


#define BTBSIZE 16 /* ADDED */

/* First we declare all the procedures */

int emulsetup(void);
void catch_exception(unsigned int instr_addr);
void initialise(unsigned int start_address, int initial_mode);
void execute(unsigned int op_code);

/* ARM execute */
void data_op(unsigned int op_code);
void clz(unsigned int op_code);
void transfer(unsigned int op_code);
void transfer_sbhw(unsigned int op_code);
void multiple(unsigned int op_code);
void branch(unsigned int op_code);
void coprocessor(unsigned int op_code);
void my_system(unsigned int op_code);
void undefined();
void breakpoint();

void mrs(unsigned int op_code);
void msr(unsigned int op_code);
void bx(unsigned int Rm, int link);
void my_multi(unsigned int op_code);
void swap(unsigned int op_code);
void normal_data_op(unsigned int op_code, int operation);
void ldm(int mode, int Rn, int reg_list, int write_back, int hat);
void stm(int mode, int Rn, int reg_list, int write_back, int hat);

int transfer_offset(int op2, int add, int imm, int sbhw);

int b_reg(int op2, int *cf);
int b_immediate(int op2, int *cf);

int bit_count(unsigned int source, int *first);

int check_cc(int condition);
void state();
void print_flags(unsigned int flags);

int not(int x);
int and(int x, int y);
int or(int x, int y);
int xor(int x, int y);
int zf(int cpsr);
int cf(int cpsr);
int nf(int cpsr);
int vf(int cpsr);

void set_flags(int operation, int a, int b, int rd, int carry);
void set_NZ(unsigned int value);
void set_CF(unsigned int a, unsigned int rd, int carry);
void set_VF_ADD(int a, int b, int rd);
void set_VF_SUB(int a, int b, int rd);
int get_reg(int reg_no, int force_user); /* returns pc+4 for ARM & pc+2 for THUMB */
void put_reg(int reg_no, int value, int force_user);
int instruction_length();

unsigned int fetch();
void inc_pc();
void endian_swap(unsigned int start, unsigned int end);
int   read_mem(unsigned int address, int size, int sign, int T, int source);
void write_mem(unsigned int address, int data, int size, int T, int source);


/* ARM disassemble */
void arm_disassemble(unsigned int address, unsigned int op_code);
void dis_data_op(unsigned int address, unsigned int op_code, int cc);
void dis_transfer(unsigned int address, unsigned int op_code, int cc);
void dis_multiple(unsigned int address, unsigned int op_code, int cc);
void dis_branch(unsigned int address, unsigned int op_code, int cc);
void dis_coprocessor(unsigned int address, unsigned int op_code, int cc);
void dis_system(unsigned int address, unsigned int op_code, int cc);
void dis_dat(unsigned int address, unsigned int op_code, int cc, int operation);
void dis_mul(unsigned int address, unsigned int op_code, int cc);
void dis_swp(unsigned int address, unsigned int op_code, int cc);
void dis_shift(unsigned int op_code);
void dis_immediate(unsigned int op2);
void dis_transfer_offset(int op2, int add, int imm, int sbhw);
void dis_mrs(unsigned int address, unsigned int op_code, int cc);
void dis_msr(unsigned int address, unsigned int op_code, int cc);
void dis_bx(unsigned int address, unsigned int op_code, int cc);
void dis_sbhw(unsigned int address, unsigned int op_code, int cc); /* signed byte - hw */
void dis_cc(int cc, int *spaces);
void dis_tab(int spaces);

/* THUMB disassemble */
void thumb_disassemble(unsigned int address, unsigned int op_code);
void dis_data0(unsigned int address, unsigned int op_code);
void dis_data1(unsigned int address, unsigned int op_code);
void dis_data_transfer(unsigned int address, unsigned int op_code);
void dis_transfer0(unsigned int address, unsigned int op_code);
void dis_transfer1(unsigned int address, unsigned int op_code);
void dis_sp_pc(unsigned int address, unsigned int op_code);
void dis_lsm_b(unsigned int address, unsigned int op_code);
void dis_thumb_branch(unsigned int address, unsigned int op_code);
void print_reg_list(int i);


/* THUMB execute */
void data0(unsigned int op_code);
void data1(unsigned int op_code);
void data_transfer(unsigned int op_code);
void transfer0(unsigned int op_code);
void transfer1(unsigned int op_code);
void sp_pc(unsigned int op_code);
void lsm_b(unsigned int op_code);
void thumb_branch(unsigned int op_code);

void printchar(char c);
void case0_addr(unsigned int op_code, char *mnemonic);


int load_fpe();
void fpe_install();

void btb_insert(unsigned int addr, unsigned int instr);
int btb_allocate(int type, unsigned int addr);
void btc_flush();

void monitor_branch_init();
void monitor_branch(unsigned int instr);
unsigned int predict(unsigned int address);

void monitor_return(unsigned int address, unsigned int instr);

void monitor_pc(unsigned int address, unsigned int last);
void monitor_pc_1(unsigned int address, unsigned int last);
unsigned int prefetch(unsigned int addr, unsigned int last);

int get_number(char *ptr);
int lsl(int value, int distance, int *cf);
int lsr(unsigned int value, int distance, int *cf);
int asr(int value, int distance, int *cf);
int ror(unsigned int value, int distance, int *cf);



unsigned int getmem32(int number);
void setmem32(int number,unsigned int reg);
unsigned char execute_instruction(void);

int  board_getchararray (int, unsigned char*);
int  board_sendchararray(int, unsigned char*);
int  board_getchar           (unsigned char*);
int  board_sendchar          (unsigned char);
int  board_getb16(int*);
int  board_sendb16(int);
int  board_getb32(int*);
int  board_sendb32(int);

/*----------------------------------------------------------------------------*/

#define mem_size     0X100000                                    /* 4Mbytes */
#define RAMSIZE      0X100000                                    /* 4Mbytes */

#define reserved_mem 0X002000                                     /* 32Kbytes */
#define user_stack   (mem_size-reserved_mem)<<2
#define start_string_addr 0X00007000                            /*ARM address */

#define max_instructions 10000000

#define prefetch_depth 3

const int   FALSE = 0;
const int   TRUE = -1;

const unsigned int nf_mask        = 0X80000000;
const unsigned int zf_mask        = 0X40000000;
const unsigned int cf_mask        = 0X20000000;
const unsigned int vf_mask        = 0X10000000;
const unsigned int if_mask        = 0X00000080;
const unsigned int ff_mask        = 0X00000040;
const unsigned int mode_mask      = 0X0000001F;
const unsigned int tf_mask        = 0X00000020;                  /* THUMB bit */

const unsigned int bit_31         = 0X80000000;
const unsigned int bit_0          = 0X00000001;

const unsigned int imm_mask       = 0X02000000;      /* orginal word versions */
const unsigned int imm_hw_mask    = 0X00400000;         /* half word versions */
const unsigned int data_op_mask   = 0X01E00000;          /* ALU function code */
const unsigned int data_ext_mask  = 0X01900000;   /* To sort out CMP from MRS */
const unsigned int arith_ext      = 0X01000000; /* Poss. arithmetic extension */
const unsigned int s_mask         = 0X00100000;
const unsigned int rn_mask        = 0X000F0000;
const unsigned int rd_mask        = 0X0000F000;
const unsigned int rs_mask        = 0X00000F00;
const unsigned int rm_mask        = 0X0000000F;
const unsigned int op2_mask       = 0X00000FFF;
const unsigned int hw_mask        = 0X00000020;
const unsigned int sign_mask      = 0X00000040;

const unsigned int mul_mask       = 0X0FC000F0;
const unsigned int long_mul_mask  = 0X0F8000F0;
const unsigned int mul_op         = 0X00000090;
const unsigned int long_mul_op    = 0X00800090;
const unsigned int mul_acc_bit    = 0X00200000;
const unsigned int mul_sign_bit   = 0X00400000;
const unsigned int mul_long_bit   = 0X00800000;

const unsigned int sbhw_mask      = 0X0E000FF0;

const unsigned int swp_mask       = 0X0FB00FF0;
const unsigned int swp_op         = 0X01000090;

const unsigned int pre_mask       = 0X01000000;
const unsigned int up_mask        = 0X00800000;
const unsigned int byte_mask      = 0X00400000;
const unsigned int write_back_mask= 0X00200000;
const unsigned int load_mask      = 0X00100000;
const unsigned int byte_sign      = 0X00000080;
const unsigned int hw_sign        = 0X00008000;

const unsigned int user_mask      = 0X00400000;

const unsigned int link_mask      = 0X01000000;
const unsigned int branch_field   = 0X00FFFFFF;
const unsigned int branch_sign    = 0X00800000;

const unsigned int undef_mask     = 0X0E000010;
const unsigned int undef_code     = 0X06000010;

const  int mem_system      = 0;                    /* sources for memor y read */
const  int mem_instruction = 1;
const  int mem_data        = 2;

const  int flag_add        = 1;
const  int flag_sub        = 2;

#define user_mode  0X00000010
#define fiq_mode   0X00000011
#define irq_mode   0X00000012
#define sup_mode   0X00000013
#define abt_mode   0X00000017
#define undef_mode 0X0000001B

/*----------------------------------------------------------------------------*/

#define btb_max_size 256
#define cyclic_replace 0
#define random_replace 1
#define rand21_replace 2
#define LRU_replace    3
#define direct_mapped16  4  /* assumes btb_size = 16 */

#define stats_length 100

#define REGSIZE 65536
#define uchar unsigned char

typedef enum {        /*Board instructions unsigned char*/
 BR_NOP     = 0x0,    /*need work on*/
 BR_PING    = 0x1, 
 BR_WOT_R_U = 0x2,
 BR_RESET   = 0x4,
 BR_FR_GET  = 0x10,
 BR_FR_SET  = 0x10,
 BR_FR_FILE = 0x14,
 BR_FR_SEND = 0x15,
 
 
 BR_WOT_U_DO= 0x20,
 BR_STOP    = 0x21,
 BR_PAUSE   = 0x22,
 BR_CONTINUE= 0x23,
 
 BR_RTF_SET = 0x24,
 BR_RTF_GET = 0x25,
 
 BR_BP_WRITE= 0x30,
 BR_BP_READ = 0x31,
 BR_BP_SET  = 0x32,
 BR_BP_GET  = 0x33,
 
 BR_WP_WRITE= 0x34,
 BR_WP_READ = 0x35,
 BR_WP_SET  = 0x36,
 BR_WP_GET  = 0x37,
 
 BR_START   = 0x80
} BR_Instruction;
typedef struct {
 char cond;
 char size;
 int addra;
 int addrb;
 int dataa[2];
 int datab[2];
 } BreakElement;

/*----------------------------------------------------------------------------*/
/* Global data */



BreakElement breakpoints[32];
BreakElement watchpoints[32];

int emul_bp_flag[2];
int emul_wp_flag[2];
uchar registers[REGSIZE];
uchar memory[RAMSIZE];
uchar status;
int paused;
unsigned int steps_togo;
unsigned int steps_reset;
char runflags;
uchar rtf;



int btb_size;
int replacement_type;

unsigned int tube_address;

int r[16];
int fiq_r[7];
int irq_r[2];
int sup_r[2];
int abt_r[2];
int undef_r[2];
unsigned int cpsr;
unsigned int spsr[32];          /* Lots of wasted space - safe for any "mode" */

int can_run, run, print_out;

unsigned int exception_para[9];

int next_file_handle;
FILE *(file_handle[20]);

int count;
int cond_count;
int exec_count;

/* Thumb counters */
int f1_a1count;    /* f* & a* are thumb & ARM patterns resp. */
int f1_a2count;
int f2_a1count;
int f2_a2count;
int f3_a1count;
int f4_a1count;
int f4_a2count;
int f4_a3count;
int f4_a4count;
int f5_a2count;
int f5_a4count;
int f6_a10count;
int f7_a11count;
int f8_a13count;
int f9_a10count;
int f10_a12count;
int f11_a10count;
int f12_a1count;
int f13_a1count;
int f14_a15count;
int f15_a15count;
int f16_a19count;
int f17_a20count;
int f18_a19count;
int f19_a19count;

unsigned int btb_cam[btb_max_size];
unsigned int btb_ram[btb_max_size];
int          btb_valid[btb_max_size];
int          btc_time[btb_max_size];
int          btb_index;

int branch_count, branch_link_count, branch_nolink_count;
int branch_taken_count, branch_lt_count, branch_nlt_count;
int branch_pt, branch_pnt, branch_npt, branch_npnt;
int branch_l_pt, branch_l_pnt, branch_l_npt, branch_l_npnt;

int branch_predicted;      /* TRUE if current operation is a predicted branch */
int branch_array1[mem_size];                  /* Last time branch encountered */
int branch_array2[mem_size];                            /* Total of distances */
int branch_array3[mem_size];                   /* Number of times encountered */
int branch_stats[stats_length];
int btc_insert_count;

int branch_array4[mem_size];                              /* Marked as cached */
int should_predict;

unsigned int last_addr;
unsigned int BTC_CAM_top;
int Non_sequential, Wrapped_block, Bottom_only;


int return_count;
int return_count_unc;
int return_count_cond;
int return_CAM[20];
int return_buff_size;
int return_alloc_ptr;
int returns_found;


int non_seq, wrong;
int glob1, glob2;


int past_opc_addr[32];        /* History buffer of fetched op. code addresses */
int past_size;                                         /* Used size of buffer */
int past_opc_ptr;                                        /* Pointer into same */
int past_count;                       /* Count of hits in instruction history */

/* Thumb stuff */
int PC;
int BL_prefix, BL_address;
int next_char;
int ARM_flag;

unsigned int value;

/*----------------------------------------------------------------------------*/
/* Entry point */

int emul_bp_flag[2] = {0,-1};
int emul_wp_flag[2] = {0,-1};
int paused=0;

main()
{
 uchar c=0;
 uchar wotrustring[]={15,13,0, 1,0,0, 0, 1, 0,0,0,0, 1,0,0,0};
 struct pollfd pollfd;
 char* pointer;
 int pint;
 int size;
 int data[2];
 int temp;
 char tempchar;
 pollfd.fd=0;
 pollfd.events=POLLIN;
 emulsetup();
 
 while(1){
  while (1>poll(&pollfd, 1, 0))
        if (2 == (status>>6) && !paused) {
                status = execute_instruction();
                if (steps_togo) {
                    steps_togo--;
                    steps_reset++;
                    if (!steps_togo) status=0x40;
                    }
                }
        else poll(&pollfd, 1, -1);
    
  read(0,&c,1);
    switch(c>>6){
       case 0:
          switch(c){
            case BR_NOP:
                break;
            case BR_PING:
                write(1,"OK00",4);
                break;
            case BR_WOT_R_U:
                board_sendchararray(wotrustring[0],wotrustring+1);
                break;
            case BR_RESET:
                boardreset();
                break;
            case BR_RTF_GET:
                board_sendchar(rtf);
                break;
            case BR_RTF_SET:
                board_getchar(&rtf);
                break;
            case BR_WOT_U_DO:
                board_sendchar(status);
                board_sendb32(steps_togo);
                board_sendb32(steps_reset);
                break;
            case BR_STOP:
                status=0x40;
                steps_togo=0;
                paused=0;
                break;
            case BR_PAUSE:
                paused=1;
                break;
            case BR_CONTINUE:
                status=0x80;
                paused=0;
                break;
            case BR_BP_GET:
                board_sendb32(emul_bp_flag[0]);
                board_sendb32(emul_bp_flag[1]);
                break;
            case BR_BP_SET:
                board_getb32(&data[0]);
                board_getb32(&data[1]);
                temp=data[1] & ~data[0];
                emul_bp_flag[0] &= ~temp;
                emul_bp_flag[1] |= temp;
                temp = data[0] & emul_bp_flag[0];
                emul_bp_flag[1] = (emul_bp_flag[1] & ~temp)
                                    | (data[1] & temp);
                break;
            case BR_BP_READ:
                board_getchar(&tempchar);
                temp=tempchar;
                board_sendchar(breakpoints[temp].cond);
                board_sendchar(breakpoints[temp].size);
                board_sendb32(breakpoints[temp].addra);
                board_sendb32(breakpoints[temp].addrb);
                board_sendb32(breakpoints[temp].dataa[0]);
                board_sendb32(breakpoints[temp].dataa[1]);
                board_sendb32(breakpoints[temp].datab[0]);
                board_sendb32(breakpoints[temp].datab[1]);
                break;                
            case BR_BP_WRITE:
                board_getchar(&tempchar);
                temp=tempchar;
                board_getchar(&breakpoints[temp].cond);
                board_getchar(&breakpoints[temp].size);
                board_getb32(&breakpoints[temp].addra);
                board_getb32(&breakpoints[temp].addrb);
                board_getb32(&breakpoints[temp].dataa[0]);
                board_getb32(&breakpoints[temp].dataa[1]);
                board_getb32(&breakpoints[temp].datab[0]);
                board_getb32(&breakpoints[temp].datab[1]);
                temp = (1<<temp) & ~emul_bp_flag[0];
                emul_bp_flag[0] |= temp;
                emul_bp_flag[1] |= temp;
                break;                
            case BR_WP_GET:
                board_sendb32(emul_wp_flag[0]);
                board_sendb32(emul_wp_flag[1]);
                break;
            case BR_WP_SET:
                board_getb32(&data[0]);
                board_getb32(&data[1]);
                temp=data[1] & ~data[0];
                emul_wp_flag[0] &= ~temp;
                emul_wp_flag[1] |= temp;
                temp = data[0] & emul_wp_flag[0];
                emul_wp_flag[1] = (emul_wp_flag[1] & ~temp)
                                    | (data[1] & temp);
                break;
            case BR_WP_READ:
                board_getchar(&tempchar);
                temp=tempchar;
                board_sendchar(watchpoints[temp].cond);
                board_sendchar(watchpoints[temp].size);
                board_sendb32 (watchpoints[temp].addra);
                board_sendb32 (watchpoints[temp].addrb);
                board_sendb32 (watchpoints[temp].dataa[0]);
                board_sendb32 (watchpoints[temp].dataa[1]);
                board_sendb32 (watchpoints[temp].datab[0]);
                board_sendb32 (watchpoints[temp].datab[1]);
                break;                
            case BR_WP_WRITE:
                board_getchar(&tempchar);
                temp=tempchar;
                board_getchar(&watchpoints[temp].cond);
                board_getchar(&watchpoints[temp].size);
                board_getb32(&watchpoints[temp].addra);
                board_getb32(&watchpoints[temp].addrb);
                board_getb32(&watchpoints[temp].dataa[0]);
                board_getb32(&watchpoints[temp].dataa[1]);
                board_getb32(&watchpoints[temp].datab[0]);
                board_getb32(&watchpoints[temp].datab[1]);
                temp = 1<<temp & ~emul_wp_flag[0];
                emul_wp_flag[0] |= temp;
                emul_wp_flag[1] |= temp;
                break;                
            default:
                break;
            }
          break;
       case 1:
        board_getb32(&size);
        if (c>>4&1) {
                pint = size & (REGSIZE-1);
                board_getb16(&size);
                while (size--)
                        if (c&8) board_sendb32(r[pint++]); 
                        else board_getb32(&r[pint++]);
                }
                
        else {
                pointer = memory    + (size     &(RAMSIZE-1));
                board_getb16(&size);
                size *= 1<<(c&7);
                if (((uchar*)pointer+size)>((uchar*)memory+RAMSIZE)) pointer -= RAMSIZE;
                if (c&8) board_sendchararray(size, pointer);
                else board_getchararray(size, pointer);
                }
        break;
       case 2:
        runflags = c;
        status=0x80;
        board_getb32(&steps_togo);
        break;
    }
//  afterinstruction(c);
  }
}


int board_getchar(unsigned char *to_get) /* get 1 char from board*/
{
 return board_getchararray(1,  to_get);
}

int board_sendchar(unsigned char to_send) /*send 1 char to the board*/
{
return board_sendchararray(1, &to_send);
}


int board_getb16(int* toget) /* get 16 bit int from board*/
{
 unsigned char from_board[2];
 if (2 != board_getchararray(2, from_board))     return 0;
 *toget=from_board[0]+(from_board[1]<<8);
 return 2;
}

int board_sendb16(int tosend) /* send 16 bit int to the board*/
{
 if (1 != board_sendchar((unsigned char) tosend & 255))
                                                 return 0;
 if (1 != board_sendchar((unsigned char) (tosend >> 8) & 255))
                                                 return 0;
 return 2;
}
		   
int board_sendb32(int tosend) /* get 32 bit int from board*/
{
 int x=4;
 while(x--){
   if (1 != board_sendchar((unsigned char) tosend & 255))
                                                 return 0;
   tosend = tosend>>8;
   }
 return 4;
}

int board_getb32(int* toget) /* send 32 bit int to the board*/
{
 int x=4;
 unsigned char temp;
 *toget=0;
 
for (x=0;x<4;x++){
   if (1 != board_getchar(&temp))            return 0;
   *toget= *toget | temp<<(x*8);
   }
   
 return 4;
}





int board_getchararray(int char_number, unsigned char* data_ptr)
{
 int ret=char_number; /* get char array from the board*/
 int replycount=0;
 int temp;
 struct pollfd pollfd;
 pollfd.fd=0;
 pollfd.events=POLLIN;
	while (char_number){
        if(!poll(&pollfd, 1, -1)) return ret-char_number;
        replycount = read(0, data_ptr, char_number);
		if (replycount<0) replycount=0;
		char_number -= replycount;
        data_ptr += replycount;
		}
	
 return ret;
}

int board_sendchararray(int char_number, unsigned char* data_ptr)
{
 write(1, data_ptr, char_number);
 return char_number; /* send char array to the board*/
}











int emulsetup(void)
{
char *test_file_name;
FILE *fp, *fopen();
char c;

int address, op_code;
int initial_mode;

glob1 = 0;
glob2 = 0;

  {
  int i;
  for (i=0;i<32;i++) past_opc_addr[i]=1;            /* Illegal op. code address */
  past_opc_ptr = 0;
  past_count   = 0;
  past_size    = 4;
  }

initial_mode  = user_mode;
can_run = TRUE;
print_out = TRUE;
btb_size = BTBSIZE;
replacement_type = random_replace;

BTC_CAM_top = 0XFFFFFFFC;
  next_file_handle = 1;

  initialise(0, initial_mode);
  monitor_branch_init();

  Non_sequential = 0;
  Wrapped_block  = 0;
  Bottom_only    = 0;

  return_count      = 0;
  return_count_unc  = 0;
  return_count_cond = 0;
  return_buff_size  = 10;
  return_alloc_ptr  = 0;
  {int i; for(i=0;i<return_buff_size;i++) return_CAM[i] = 0;}
  returns_found     = 0;

  non_seq=0; wrong=0;

  run = can_run;

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
}
uchar execute_instruction(void)
{
  int instr;
  int instr_addr;

  f1_a1count   = 0;  f1_a2count   = 0;  f2_a1count   = 0;  f2_a2count   = 0;
  f3_a1count   = 0;  f4_a1count   = 0;  f4_a2count   = 0;  f4_a3count   = 0;
  f4_a4count   = 0;  f5_a2count   = 0;  f5_a4count   = 0;  f6_a10count  = 0;
  f7_a11count  = 0;  f8_a13count  = 0;  f9_a10count  = 0;  f10_a12count = 0;
  f11_a10count = 0;  f12_a1count  = 0;  f13_a1count  = 0;  f14_a15count = 0;
  f15_a15count = 0;  f16_a19count = 0;  f17_a20count = 0;  f18_a19count = 0;
  f19_a19count = 0;

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

    instr_addr = get_reg(15, FALSE) - instruction_length();

//    if (instr_addr <= 0X00000020) catch_exception(instr_addr);                 /* Assume operating system */

    last_addr = get_reg(15, FALSE) - instruction_length();

    instr = fetch();                                                 /* FETCH */

    if ((cpsr & tf_mask) == 0)              /* Count conditional instructions */
      {if ((instr & 0XF0000000) != 0XE0000000) cond_count++;}     /* ARM case */
    else
      {                                                         /* Thumb case */
      if (((instr & 0XF000) == 0XD000) && ((instr & 0X0F00) != 0X0F00))
        cond_count++;
      }
  execute(instr);                                                /* EXECUTE */
  return 0x80;
}

/*----------------------------------------------------------------------------*/

void catch_exception(unsigned int instr_addr)
{
/*
if (instr_addr == 0X00000004)
  {
  printf("\n** Undefined instruction trap **\n\n");
  }
*/

if ((instr_addr <= 0X00000020) && (instr_addr != 0X00000004))
                  /* Ignore "undefined" entries for now @@ */
  {
  unsigned int i, offset, handler_addr, sp;

  printf("Exception entry\n");
  state();

  offset = read_mem(instr_addr, 4, FALSE, FALSE, mem_system) & branch_field;
  if ((offset & branch_sign) != 0)
    offset = offset | ~branch_field;                           /* sign extend */

  handler_addr = (get_reg(15, FALSE)+4) + (offset<<2);

            printf("initial PC: %08x\n", (get_reg(15, FALSE) + 4));

  i = instr_addr >> 2;
  sp = get_reg(13, FALSE) >> 2;                            /* Stack registers */
  write_mem(--sp, get_reg(14, FALSE), 4, FALSE, mem_system);
  write_mem(--sp, get_reg(12, FALSE), 4, FALSE, mem_system);
  write_mem(--sp, get_reg(11, FALSE), 4, FALSE, mem_system);
  write_mem(--sp, get_reg(10, FALSE), 4, FALSE, mem_system);
  put_reg(13, sp<<2, FALSE);
  put_reg(11, exception_para[i], FALSE);                      /* i = PC value */

  switch (i)
    {
    case 2:                                                            /* SWI */
      put_reg(10, read_mem((get_reg(14, FALSE)-4)>>2, 4, FALSE,
                            FALSE, mem_system) & 0X00FFFFFF, FALSE);
                                                        /* Recover SWI number */
      break;

    case 8:     /*
                                                    /* Needs error pointer @@ */
      break;
    }

    printf("Entering handler @ %08X\n\n", handler_addr);
    put_reg(15, handler_addr, FALSE);                             /* Despatch */
  }

return;
}

/*----------------------------------------------------------------------------*/
boardreset()
{
initialise(0, user_mode);
}



void initialise(unsigned int start_address, int initial_mode)
{
int i;

cpsr = 0X000000C0 | initial_mode;                       /* Disable interrupts */
                                                     /* in ARM mode T-bit = 0 */
for (i=0;i<15;i++) r[i] = 0;

r[15] = start_address;
r[13]      = user_stack;                         /* Initialise stack pointers */
fiq_r[5]   = 0X00000400;
irq_r[0]   = 0X00000500;
sup_r[0]   = 0X00000800;
abt_r[0]   = 0X00000700;
undef_r[0] = 0X00000600;

return;
}

/*----------------------------------------------------------------------------*/

void execute(unsigned int op_code)
{

/* ARM or THUMB ? */

if ((cpsr & tf_mask) != 0)                                             /* Thumb */
  {
  inc_pc();

  op_code = op_code & 0XFFFF;

  /* Conditonals */
  if (((op_code & 0XF000) != 0XD000) || ((op_code & 0X0F00) == 0X0F00)
   || (check_cc(op_code>>8) == TRUE)) exec_count++;

  switch (op_code & 0XE000)
    {
    case 0X0000: data0(op_code);         break;
    case 0X2000: data1(op_code);         break;
    case 0X4000: data_transfer(op_code); break;
    case 0X6000: transfer0(op_code);     break;
    case 0X8000: transfer1(op_code);     break;
    case 0XA000: sp_pc(op_code);         break;
    case 0XC000: lsm_b(op_code);         break;
    case 0XE000: thumb_branch(op_code);  break;
    }
  }
else
  {
  inc_pc();                                         /* Easier here than later */

  if ((check_cc(op_code>>28) == TRUE)                      /* Check condition */
  || ((op_code & 0XFE000000) == 0XFA000000))      /* Nasty non-orthogonal BLX */
    {
    exec_count++;

    switch ((op_code>>25) & 0X00000007)
      {
      case 0X0: data_op(op_code);     break;    /* includes load/store hw & sb*/
      case 0X1: data_op(op_code);     break;/* data processing & MSR immediate*/
      case 0X2: transfer(op_code);    break;
      case 0X3: transfer(op_code);    break;
      case 0X4: multiple(op_code);    break;
      case 0X5: branch(op_code);      break;
      case 0X6: coprocessor(op_code); break;
      case 0X7: my_system(op_code);   break;
      }
    }
  else
    if (print_out == TRUE) printf("Skipping %08X\n", op_code);
  }

return;
}

/*----------------------------------------------------------------------------*/

int is_it_sbhw(unsigned int op_code)
{
if (((op_code & 0X0E000090) == 0X00000090)
 && ((op_code & 0X00000060) != 0X00000000)                  /* No multiplies */
 && ((op_code & 0X00100040) != 0X00000040))              /* No signed stores */
  {
  if (((op_code & 0X00400000) != 0) || ((op_code & 0X00000F00) == 0))
    return TRUE;
  else
    return FALSE;
  }
else
  return FALSE;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void data_op(unsigned int op_code)
{
int operation;
int shift_carry;

if (((op_code & mul_mask) == mul_op) ||
    ((op_code & long_mul_mask) == long_mul_op)) my_multi(op_code);
else
  if (is_it_sbhw(op_code) == TRUE) transfer_sbhw(op_code);
  else
    if ((op_code & swp_mask) == swp_op) swap(op_code);
    else
      {
      operation = (op_code & data_op_mask)>>21;

           /* TST, TEQ, CMP, CMN - all lie in following range, but have S set */
      if ((op_code & data_ext_mask) == arith_ext)      /* PSR transfers OR BX */
        {
        if ((op_code & 0X0FBF0FFF) == 0X010F0000) mrs(op_code);        /* MRS */
        else
          if (((op_code & 0X0DB6F000) == 0X0120F000)
           && ((op_code & 0X02000010) != 0X00000010)) msr(op_code);    /* MSR */
          else
            if ((op_code & 0X0FFFFFD0) == 0X012FFF10)               /* BX/BLX */
              bx(op_code & rm_mask, op_code & 0X00000020);
            else
              if ((op_code & 0XFFF000F0) == 0XE1200070)         /* Breakpoint */
                breakpoint();
              else
                if ((op_code & 0X0FFF0FF0) == 0X016F0F10) clz(op_code);/* CLZ */
                else undefined();
        }
      else
        normal_data_op(op_code, operation); /* All data processing operations */
      }
return;
}

/*----------------------------------------------------------------------------*/

void transfer_sbhw(unsigned int op_code)
{
unsigned int address;
int size, sign;
int offset, rd;

switch (op_code & 0X00000060)
  {
  case 0X00: printf("Multiply shouldn't be here!\n"); break;        /* Error! */
  case 0X20: size = 2; sign = FALSE; break;                              /* H */
  case 0X40: size = 1; sign = TRUE;  break;                             /* SB */
  case 0X60: size = 2; sign = TRUE;  break;                             /* SH */
  }

rd = ((op_code & rd_mask) >> 12);

address = get_reg(((op_code & rn_mask)>>16), FALSE);

offset = transfer_offset(op_code & op2_mask, op_code & up_mask,
                         op_code & imm_hw_mask, TRUE);

if ((op_code & pre_mask) != 0) address = address + offset;       /* pre-index */

if ((op_code & load_mask) == 0)                                      /* store */
  write_mem(address, get_reg(rd, FALSE), size, FALSE, mem_data);
else                                                                  /* load */
  put_reg(rd, read_mem(address, size, sign, FALSE, mem_data), FALSE);
                                                                /* post index */

if ((op_code & pre_mask) == 0)                    /* post index with writeback*/
  put_reg((op_code & rn_mask) >> 16, address + offset, FALSE);
else
  if ((op_code & write_back_mask) != 0)
    put_reg((op_code & rn_mask) >> 16, address, FALSE);

return;
}

/*----------------------------------------------------------------------------*/

void mrs(unsigned int op_code)
{
if ((op_code & 0X00400000) == 0)
  put_reg((op_code & rd_mask)>>12, cpsr, FALSE);
else
  put_reg((op_code & rd_mask)>>12, spsr[cpsr & mode_mask], FALSE);

return;
}

/*----------------------------------------------------------------------------*/

void msr(unsigned int op_code)
{
int mask, source;

switch (op_code & 0X00090000)
  {
  case 0X00000000: mask = 0X00000000; break;
  case 0X00010000: mask = 0X0FFFFFFF; break;
  case 0X00080000: mask = 0XF0000000; break;
  case 0X00090000: mask = 0XFFFFFFFF; break;
  }
if ((cpsr & mode_mask) == 0X10) mask = mask & 0XF0000000;        /* User mode */

if ((op_code & imm_mask) == 0)                 /* Test applies for both cases */
  source = get_reg(op_code & rm_mask, FALSE) & mask;
else
  {
  unsigned int x, y, dummy;

  x = op_code & 0X0FF;                                     /* Immediate value */
  y = (op_code & 0XF00)>>7;                              /* Number of rotates */
  source = ((x >> y) | lsl(x, 32-y, &dummy)) & mask;
  }

if ((op_code & 0X00400000) == 0)
  cpsr = (cpsr & ~mask) | source;
else
  spsr[cpsr & mode_mask] =  (spsr[cpsr & mode_mask] & ~mask) | source;

return;
}

/*----------------------------------------------------------------------------*/

void bx(unsigned int Rm, int link) /* Link is performed if "link" is NON-ZERO */
{
int PC, offset;
int t_bit;

PC = get_reg(15, FALSE);
if ((cpsr & tf_mask) != 0) {PC = PC - 2; PC = PC | 1;} /* Remember Thumb mode */
else                        PC = PC - 4;

offset = get_reg(Rm, FALSE) & 0XFFFFFFFE;
t_bit  = get_reg(Rm, FALSE) & 0X00000001;

if (t_bit == 1) cpsr = cpsr |  tf_mask;
else            cpsr = cpsr & ~tf_mask;

put_reg(15, offset, FALSE);                                      /* Update PC */
if (link != 0) put_reg(14, PC, FALSE);                         /* Link if BLX */

return;
}

/*----------------------------------------------------------------------------*/

void my_multi(unsigned int op_code)
{
int acc;

if ((op_code & mul_long_bit) == 0)                                  /* Normal */
  {
  acc = get_reg(op_code & rm_mask, FALSE)
      * get_reg((op_code & rs_mask)>>8, FALSE);

  if ((op_code & mul_acc_bit) != 0)
    acc = acc + get_reg((op_code & rd_mask)>>12, FALSE);

  put_reg((op_code & rn_mask)>>16, acc, FALSE);

  if ((op_code & s_mask) != 0) set_NZ(acc);                          /* Flags */
  }
else                                                                  /* Long */
  {
  unsigned int Rm, Rs, th, tm, tl;
  int sign;

  Rm = get_reg( op_code & rm_mask,       FALSE);
  Rs = get_reg((op_code & rs_mask) >> 8, FALSE);

  sign = 0;
  if ((op_code & mul_sign_bit) != 0)                                /* Signed */
    {
    if ((Rm & bit_31) != 0) {Rm = ~Rm + 1; sign = 1;}
    if ((Rs & bit_31) != 0) {Rs = ~Rs + 1; sign = sign ^ 1;}
    }                                            /* Everything now `positive' */

  tl = (Rm & 0X0000FFFF) * (Rs & 0X0000FFFF);
  th = ((Rm>>16) & 0X0000FFFF) * ((Rs>>16) & 0X0000FFFF);
  tm = ((Rm>>16) & 0X0000FFFF) * (Rs & 0X0000FFFF);
  Rm = ((Rs>>16) & 0X0000FFFF) * (Rm & 0X0000FFFF);    /* Rm no longer needed */
  tm = tm + Rm; if (tm < Rm) th = th + 0X00010000;         /* Propagate carry */
  tl = tl + (tm<<16); if (tl < (tm<<16)) th = th + 1;
  th = th + ((tm>>16) & 0X0000FFFF);

  if (sign != 0)                                     /* Change sign of result */
    {th = ~th; tl = ~tl + 1; if (tl = 0) th = th + 1;}

  if ((op_code & mul_acc_bit) != 0)
    {
    tm = tl + get_reg((op_code & rd_mask)>>12, FALSE);
    if (tm < tl) th = th + 1;                              /* Propagate carry */
    tl = tm;
    th = th + get_reg((op_code & rn_mask)>>16, FALSE);
    }

  put_reg((op_code & rd_mask)>>12, tl, FALSE);
  put_reg((op_code & rn_mask)>>16, th, FALSE);

  if ((op_code & s_mask) != 0)
    set_NZ(th | (((tl >> 16) | tl) & 0X0000FFFF));                   /* Flags */
  }

return;
}

/*----------------------------------------------------------------------------*/

void swap(unsigned int op_code)
{
unsigned int address, data, size;

address = get_reg((op_code & rn_mask)>>16, FALSE);
if ((op_code & byte_mask) != 0) size = 1;
else                            size = 4;

data = read_mem(address, size, FALSE, FALSE, mem_data);
write_mem(address, get_reg(op_code & rm_mask, FALSE),
          size, FALSE, mem_data);
put_reg((op_code & rd_mask)>>12, data, FALSE);

return;
}

/*----------------------------------------------------------------------------*/

void normal_data_op(unsigned int op_code, int operation)
{
int rd, a, b, mode;
int shift_carry;
int CPSR_special;

mode = cpsr & mode_mask;
CPSR_special = FALSE;
shift_carry = 0;
a = get_reg((op_code & rn_mask)>>16, FALSE); /* force_user = FALSE */

if ((op_code & imm_mask) == 0)
  b = b_reg(op_code & op2_mask, &shift_carry);
else
  b = b_immediate(op_code & op2_mask, &shift_carry);

switch (operation)                                               /* R15s @@?! */
  {
  case 0X0: rd = a & b; break;                                         /* AND */
  case 0X1: rd = a ^ b; break;                                         /* EOR */
  case 0X2: rd = a - b; break;                                         /* SUB */
  case 0X3: rd = b - a; break;                                         /* RSB */
  case 0X4: rd = a + b; break;                                         /* ADD */
  case 0X5: rd = a + b; if ((cpsr & cf_mask) != 0) rd = rd + 1; break; /* ADC */
  case 0X6: rd = a - b - 1; if ((cpsr & cf_mask)!=0) rd = rd+1; break; /* SBC */
  case 0X7: rd = b - a - 1; if ((cpsr & cf_mask)!=0) rd = rd+1; break; /* RSC */
  case 0X8: rd = a & b; break;                                         /* TST */
  case 0X9:                                                            /* TEQ */
    rd = a ^ b;
    if ((op_code & rd_mask) == 0XF000)                                /* TEQP */
      {
      CPSR_special = TRUE;
      if (mode != user_mode) cpsr = spsr[mode];
      }
    break;
  case 0XA: rd = a - b; break;                                         /* CMP */
  case 0XB: rd = a + b; break;                                         /* CMN */
  case 0XC: rd = a | b; break;                                         /* ORR */
  case 0XD: rd = b; break;                                             /* MOV */
  case 0XE: rd = a & ~b; break;                                        /* BIC */
  case 0XF: rd = ~b; break;                                            /* MVN */
  }

if ((operation & 0XC) != 0X8) put_reg((op_code & rd_mask)>>12, rd, FALSE);
                                            /* Return result unless a compare */

if (((op_code & s_mask) != 0) && (CPSR_special != TRUE))             /* S-bit */

  {                                                    /* Want to change CPSR */
  if (((op_code & rd_mask)>>12) == 0XF)                       /* PC and S-bit */
    {
    if (mode != user_mode) cpsr = spsr[mode];           /* restore saved CPSR */
    else printf("SPSR_user read attempted\n");
    }
  else                                               /* other dest. registers */
    {
    switch (operation)
      {                                                           /* LOGICALs */
      case 0X0:                                                        /* AND */
      case 0X1:                                                        /* EOR */
      case 0X8:                                                        /* TST */
      case 0X9:                                                        /* TEQ */
      case 0XC:                                                        /* ORR */
      case 0XD:                                                        /* MOV */
      case 0XE:                                                        /* BIC */
      case 0XF:                                                        /* MVN */
        set_NZ(rd);
        if (shift_carry == TRUE) cpsr = cpsr |  cf_mask;      /* CF := output */
        else                     cpsr = cpsr & ~cf_mask;      /* from shifter */
        break;

      case 0X2:                                                        /* SUB */
      case 0XA:                                                        /* CMP */
        set_flags(flag_sub, a, b, rd, 1);
        break;

      case 0X6:                        /* SBC - Needs testing more !!!   @@@@ */
        set_flags(flag_sub, a, b, rd, cpsr & cf_mask);
        break;

      case 0X3:                                                        /* RSB */
        set_flags(flag_sub, b, a, rd, 1);
        break;

      case 0X7:                                                        /* RSC */
        set_flags(flag_sub, b, a, rd, cpsr & cf_mask);
        break;

      case 0X4:                                                        /* ADD */
      case 0XB:                                                        /* CMN */
        set_flags(flag_add, a, b, rd, 0);
        break;

      case 0X5:                                                        /* ADC */
        set_flags(flag_add, a, b, rd, cpsr & cf_mask);
        break;
      }
    }
  }

return;
}

/*----------------------------------------------------------------------------*/
/* shift type: 00 = LSL, 01 = LSR, 10 = ASR, 11 = ROR */

int b_reg(int op2, int *cf)
{
unsigned int shift_type, reg, distance, result;

reg = get_reg(op2 & 0X00F, FALSE);                                /* Register */
shift_type = (op2 & 0X060)>>5;                               /* Type of shift */
if ((op2 & 0X010) == 0)
  {                                                        /* Immediate value */
  distance = (op2 & 0XF80)>>7;
  if (distance == 0)                                         /* Special cases */
    {
    if (shift_type == 3)
      {
      shift_type = 4;                                                  /* RRX */
      distance = 1;                                     /* Something non-zero */
      }
    else
      if (shift_type != 0) distance = 32;                     /* LSL excluded */
    }
  }
else
  distance = (get_reg((op2 & 0XF00)>>8, FALSE) & 0XFF);     /* Register value */

  *cf = ((cpsr & cf_mask) != 0);                            /* Previous carry */
  switch (shift_type)
    {
    case 0X0: result = lsl(reg, distance, cf); break;                  /* LSL */
    case 0X1: result = lsr(reg, distance, cf); break;                  /* LSR */
    case 0X2: result = asr(reg, distance, cf); break;                  /* ASR */
    case 0X3: result = ror(reg, distance, cf); break;                  /* ROR */
    case 0X4:                                                       /* RRX #1 */
      result = reg >> 1;
      if ((cpsr & cf_mask) == 0) result = result & ~bit_31;
      else                       result = result |  bit_31;
      *cf = ((reg & bit_0) != 0);
      break;
    }

if (*cf) *cf=TRUE; else *cf=FALSE;                     /* Change to "Boolean" */
return result;
}

/*----------------------------------------------------------------------------*/

int b_immediate(int op2, int *cf)
{
unsigned int x, y, dummy;

x = op2 & 0X0FF;                                           /* Immediate value */
y = (op2 & 0XF00)>>7;                                    /* Number of rotates */
if (y == 0) *cf = ((cpsr & cf_mask) != 0);                  /* Previous carry */
else        *cf = (((x >> (y-1)) & bit_0) != 0);
if (*cf) *cf=TRUE; else *cf=FALSE;                     /* Change to "Boolean" */
return ror(x, y, &dummy);                                /* Circular rotation */
}

/*----------------------------------------------------------------------------*/

void clz(unsigned int op_code)
{
int i, j;

i = 31;
j = get_reg(op_code & rm_mask, FALSE);

if (j == 0) i = 32;
else {i = 31; while ((j & 0X80000000) == 0) {i--; j = j<<1;}}

put_reg((op_code & rd_mask)>>12, i, FALSE);

return;
}

/*----------------------------------------------------------------------------*/

void transfer(unsigned int op_code)
{
unsigned int address;
int offset, rd, size, T;

if ((op_code & undef_mask) == undef_code) undefined();
else
  {
  if ((op_code & byte_mask) == 0) size = 4;
  else                            size = 1;

  if (((op_code & pre_mask) == 0)
   && ((op_code & write_back_mask) != 0)) T = TRUE;
  else                                    T = FALSE;

  rd = (op_code & rd_mask)>>12;

  address = get_reg((op_code & rn_mask)>>16, FALSE);

  offset = transfer_offset(op_code & op2_mask,
                           op_code & up_mask,
                           op_code & imm_mask,           /* bit(25) = 1 -> reg */
                           FALSE);

  if ((op_code & pre_mask) != 0) address = address + offset;      /* Pre-index */

  if ((op_code & load_mask) == 0)
    write_mem(address, get_reg(rd, FALSE), size, T, mem_data);
  else
    put_reg(rd, read_mem(address, size, FALSE, T, mem_data), FALSE);

  if ((op_code & pre_mask) == 0)                                /* Post-index */
    put_reg((op_code & rn_mask)>>16, address + offset, FALSE);
                                                     /* Post index write-back */
  else
    if ((op_code & write_back_mask) != 0)
      put_reg((op_code & rn_mask)>>16, address, FALSE);
                                                      /* Pre index write-back */
  }
return;
}

/*----------------------------------------------------------------------------*/

int transfer_offset(int op2, int add, int imm, int sbhw)
{                                   /* add and imm are zero/non-zero Booleans */
int offset;
    int cf;                                                /* dummy parameter */

if (sbhw == FALSE)                                       /* Addressing mode 2 */
  {
  if (imm != 0) offset = b_reg(op2, &cf);               /* bit(25) = 1 -> reg */
  else          offset = op2 & 0XFFF;
  }
else                                                     /* Addressing mode 3 */
  {
  if (imm != 0) offset = ((op2 & 0XF00) >> 4) | (op2 & 0X00F);
  else          offset = b_reg(op2, &cf);
  }

if (add == 0) offset = -offset;

return offset;
}

/*----------------------------------------------------------------------------*/

void multiple(unsigned int op_code)
{

if ((op_code & load_mask) == 0)
  stm((op_code & 0X01800000)>>23, (op_code & rn_mask)>>16, op_code & 0X0000FFFF,
       op_code & write_back_mask,  op_code & user_mask);
else
  ldm((op_code & 0X01800000)>>23, (op_code & rn_mask)>>16, op_code & 0X0000FFFF,
       op_code & write_back_mask,  op_code & user_mask);

return;
}

/*----------------------------------------------------------------------------*/

int bit_count(unsigned int source, int *first)
{
int count, reg;

count = 0; reg = 0; *first = -1;

while (source != 0)
  {
  if ((source & bit_0) != 0)
    {
    count = count + 1;
    if (*first < 0) *first = reg;
    }
  source = source >> 1;
  reg = reg + 1;
  }

return count;
}

/*----------------------------------------------------------------------------*/

void ldm(int mode, int Rn, int reg_list, int write_back, int hat)
{                           /* Last two parameters are zero/non-zero Booleans */

int address, new_base, count, first_reg, reg, data;
int force_user;
int r15_inc;                                            /* internal `Boolean' */

address  = get_reg(Rn, FALSE);
count    = bit_count(reg_list, &first_reg);
r15_inc  = (reg_list & 0X00008000) != 0;                       /* R15 in list */

switch (mode)
  {
  case 0: new_base = address - 4 * count; address = new_base + 4; break;
  case 1: new_base = address + 4 * count;                         break;
  case 2: new_base = address - 4 * count; address = new_base;     break;
  case 3: new_base = address + 4 * count; address = address + 4;  break;
  }

address = address & 0XFFFFFFFC;           /* Bottom 2 bits ignored in address */

if (write_back != 0) put_reg(Rn, new_base, FALSE);

if ((hat != 0) && !r15_inc) force_user = TRUE;
else force_user = FALSE;                     /* Force user unless R15 in list */

reg = 0;

while (reg_list != 0)
  {
  if ((reg_list & bit_0) != 0)
    {
    data = read_mem(address, 4, FALSE, FALSE, mem_data);    /* Keep for later */
    put_reg(reg, data, force_user);
    address = address + 4;
    }
  reg_list = reg_list >> 1;
  reg = reg + 1;
  }

if (r15_inc)                                                   /* R15 in list */
  {
  if ((data & 1) != 0) cpsr = cpsr |  tf_mask;/* data left over from last load*/
  else                 cpsr = cpsr & ~tf_mask; /* used to set instruction set */
  if (hat != 0) cpsr = spsr[cpsr & mode_mask];        /* ... and if S bit set */
  }                                                        /*  @@@ untested */

return;
}

/*----------------------------------------------------------------------------*/

void stm(int mode, int Rn, int reg_list, int write_back, int hat)
{                           /* Last two parameters are zero/non-zero Booleans */

int address, new_base, count, first_reg, reg;
int force_user, special;

address  = get_reg(Rn, FALSE);
count    = bit_count(reg_list, &first_reg);

switch (mode)
  {
  case 0: new_base = address - 4 * count; address = new_base + 4; break;
  case 1: new_base = address + 4 * count;                         break;
  case 2: new_base = address - 4 * count; address = new_base;     break;
  case 3: new_base = address + 4 * count; address = address + 4;  break;
  }

address = address & 0XFFFFFFFC;           /* Bottom 2 bits ignored in address */

special = FALSE;
if (write_back != 0)
  {
  if (Rn == first_reg) special = TRUE;
  else                 put_reg(Rn, new_base, FALSE);
  }

if (hat != 0) force_user = TRUE; else force_user = FALSE;

reg = 0;

while (reg_list != 0)
  {
  if ((reg_list & bit_0) != 0)
    {
    write_mem(address, get_reg(reg, force_user), 4, FALSE, mem_data);
    address = address + 4;
    }
  reg_list = reg_list >> 1;
  reg = reg + 1;
  }

if (special == TRUE) put_reg(Rn, new_base, FALSE);

return;
}

/*----------------------------------------------------------------------------*/

void branch(unsigned int op_code)
{
int offset, PC;

PC = get_reg(15, FALSE);                 /* Get this now in case mode changes */

if (((op_code & link_mask) != 0) || ((op_code & 0XF0000000) == 0XF0000000))
  put_reg(14, get_reg(15, FALSE) - 4, FALSE);

offset = (op_code & branch_field) << 2;
if ((op_code & branch_sign) != 0)
  offset = offset | (~(branch_field<<2) & 0XFFFFFFFC);         /* sign extend */

if ((op_code & 0XF0000000) == 0XF0000000)                 /* Other BLX fix-up */
  {
  offset = offset | ((op_code>>23) & 2);
  cpsr = cpsr |  tf_mask;
  }

put_reg(15, PC + offset, FALSE);

return;
}

/*----------------------------------------------------------------------------*/

void coprocessor(unsigned int op_code)
{
/*
printf("Coprocessor %d data transfer\n", (op_code>>8) & 0XF);
*/
undefined();
return;
}

/*----------------------------------------------------------------------------*/

void my_system(unsigned int op_code)
{
int temp;

if (((op_code & 0X0F000000) == 0X0E000000)
                                     /* bodge to allow thumb to use this code */
 || ((op_code & 0X0F000000) == 0X0C000000)
 || ((op_code & 0X0F000000) == 0X0D000000))
  {                                                       /* Coprocessor op.s */
  printf("whoops -undefined \n");
/*
  if ((op_code & 0X00000010) == 0X00000000)
    printf("Coprocessor %d data operation\n", (op_code>>8) & 0XF);
  else
    printf("Coprocessor %d register transfer\n", (op_code>>8) & 0XF);
*/
  undefined();
  }
else
  {
  btc_flush();                                       /* Flush BTC on SWI call */

  if (print_out == TRUE)
    printf("\n*** SWI CALL %06X ***\n\n", (op_code & 0X00FFFFFF));

  switch (op_code & 0X00FFFFFF)
    {
    case 0:                                                /* Print character */
      temp = get_reg(0, FALSE) & 0XFF;
      if (temp == 0X04) run = FALSE;      /* Printing ^D terminates programme */
      else
        {
        if (print_out == TRUE)
          {
          if ((temp >= 0X20) && (temp != 0X7F))
            printf("Print character %02X, %c\n", temp, temp);
          else
            printf("Print character %02X, .\n", temp);
          }
        else
          switch (temp)
            {
            case 0X0A: printf("\n");
            case 0X0D: printf("\n");
            default:   printf("%c", temp);
            }
        }
      break;

    case 1:                                               /* Print string @PC */
      {
      unsigned int i;
      char c;

      printf("RISC OS print string\n");   /* Complete guess on operation */

/*
      i = get_reg(15, FALSE)-4;
*/
      i = get_reg(15, FALSE);                        /* Is this really right? */
      while ((c=read_mem(i++, 1, FALSE, FALSE, mem_system)) >= 0X02)
        {
        if ((c == 0X0A) || (c == 0X0D)) printf("\n");
        else printf("%c", c);
        }
      printf("\n");                             /* i points just after string */
      i = (i + 3) & 0XFFFFFFFC;   /* align */
      put_reg(15, i, FALSE);
      }
/*run = FALSE;                                               /* and get out @@@ */
      break;


    case 2:                                                /* Print character */
      printf("Print string\n");
      {
      int i;
      char c;

      i = get_reg(0, FALSE);
      while (c = read_mem(i++, 1, FALSE, FALSE, mem_system) != 0)
        printf("%c", c);
      printf("\n");
      }
      break;

    case 0x10:
      printf("Get Environment\n");
      put_reg(0, start_string_addr, FALSE);
      put_reg(1, (mem_size-reserved_mem-1)<<2, FALSE);
      break;

    case 0X11:
      printf("Halt emulation\n");
      run = FALSE;
      break;

    case 0X16:
      printf("Enter supervisor mode\n");
      cpsr = (cpsr & ~mode_mask) | sup_mode;
      break;

    case 0X60:
      printf("GetErrNo - returns 0\n");
      put_reg(0, 0, FALSE);
      break;

    case 0X61:
      printf("Read clock\n");
      put_reg(0, clock()/10000, FALSE);          /* SUN OS hack - clock in us */
      break;

    case 0X63:
      printf("Read time\n");
      put_reg(0, (int) time(NULL), FALSE);
      break;

    case 0X66:
      {
      unsigned int i, j;
      char c;
      char file_name[10];

      printf("Open file ");
      i = get_reg(0, FALSE);
      j = 0;
      while ((c = read_mem(i++, 1, FALSE, FALSE, mem_system)) != '\0')
        {
        file_name[j++] = c;
        printf("%c", c);
        }
      file_name[j] = '\0';
      printf(" mode ");
      switch (get_reg(1, FALSE))
        {
        case 0:  printf("r");   break;
        case 1:  printf("rb");  break;
        case 2:  printf("r+");  break;
        case 3:  printf("r+b"); break;
        case 4:  printf("w");   break;
        case 5:  printf("wb");  break;
        case 6:  printf("w+");  break;
        case 7:  printf("w+b"); break;
        case 8:  printf("a");   break;
        case 9:  printf("ab");  break;
        case 10: printf("a+");  break;
        case 11: printf("a+b"); break;
        default: printf("%d", get_reg(1, FALSE));  break;
        }
      printf("\n");

      if ((file_name[0] == ':') &&
          (file_name[1] == 't') &&
          (file_name[2] == 't') &&
          (file_name[3] == '\0'))
        {
        file_name[0] = 'f';
        file_name[1] = '0'+next_file_handle;
        file_name[2] = '\0';
        }

      printf("File name %s, internal handle %d\n", file_name, next_file_handle);

      file_handle[next_file_handle] = fopen(file_name, "r");
                                                           /* No error checks */
      if (file_handle[next_file_handle] == NULL)
        {
if (get_reg(1, FALSE) != 4)
{  /* @@@  BODGE: Don't worry about "w" files to prevent complaints if file not found */
        run = FALSE;
        printf("File %s not found; aborting\n", file_name);
}
        }
      put_reg(0, next_file_handle, FALSE);
      next_file_handle++;
      }
      break;

    case 0X68:
      printf("Close file %d\n", get_reg(0, FALSE));
      if (file_handle[get_reg(0, FALSE)] != NULL)
        fclose(file_handle[get_reg(0, FALSE)]);
      put_reg(0, 0, FALSE);
      break;

    case 0X69:
      {
      unsigned int i, j;

      printf("Write to file %d\n", get_reg(0, FALSE));
      i = get_reg(1, FALSE);
      for (j = get_reg(2, FALSE); j > 0; j--)
        {
        printf("%c", read_mem(i, 1, FALSE, FALSE, mem_system),
                     read_mem(i, 1, FALSE, FALSE, mem_system));
        i++;
        }
      put_reg(0, 0, FALSE);
      }
      break;

    case 0X6A:
      {
      unsigned int i, j, k;
      int  l;

      k = get_reg(0, FALSE);
      printf("Read from file %d\n", k);
      i = get_reg(1, FALSE);
      l = ~EOF;                              /* Not really a character */
      for (j = get_reg(2, FALSE); (j > 0) && (l != EOF); )
        {
        l = fgetc(file_handle[k]);
        if (l == EOF)
          printf("End of file\n");
        else
          {
/*
          printf("%c\n", l);
*/
          write_mem(i, l, 1, FALSE, mem_system);
          i++;
          j--;
          }
        }
      put_reg(0, j, FALSE);
      }
      break;

    case 0X6E:
      printf("IsTTY %d\n", get_reg(0, FALSE));
      put_reg(0, 0, FALSE);
      printf("(Answer 0)\n");
      break;

    case 0X70:
      {
      unsigned int old_vector, old_para;

      printf("Install Handler\n");
      printf("Exception number  %08X  ", get_reg(0, FALSE));
      printf("Value for handler %08X  ", get_reg(1, FALSE));
      printf("Handler address   %08X\n", get_reg(2, FALSE));

                             /* Refuse undefined trap 'cos FPA code will barf */
if (get_reg(0, FALSE) == 1)
  printf("I decline to do this one!\n");
else
  {
      old_vector = read_mem(get_reg(0, FALSE), 4, FALSE, FALSE, mem_system)
                && branch_field;
      if ((old_vector & branch_sign) != 0)
        old_vector = old_vector | ~branch_field;               /* sign extend */
      old_vector = (old_vector + get_reg(0, FALSE) + 2) << 2;
      old_para = exception_para[get_reg(0, FALSE)];

      write_mem(get_reg(0, FALSE)<<2,
                0XEA000000 |
               (0X00FFFFFF & ((get_reg(2, FALSE) >> 2) - 2 - get_reg(0, FALSE))),
                4, FALSE, mem_system);
                                   /* Vector translated as branch instruction */
      exception_para[get_reg(0, FALSE)] = get_reg(1, FALSE);

      put_reg(1, old_para, FALSE);
      put_reg(2, old_vector, FALSE);
  }
      }
      break;

    case 0X71:
      {
      unsigned int addr;
      char c;

      addr = get_reg(0, FALSE);
      printf("Generate Error #%08X\n",
             read_mem(addr, 4, FALSE, FALSE, mem_system));
      addr = addr + 4;
      while ((c = read_mem(addr++, 1, FALSE, FALSE, mem_system)) != 0)
        printf("%c", c);
      printf("\n");
      state();
      }
      break;

    default:
      if (print_out == TRUE)
        printf("Untrapped SWI call %06X\n", (op_code & 0X00FFFFFF));
      spsr[sup_mode] = cpsr;
      cpsr = (cpsr & ~mode_mask) | sup_mode;
      cpsr =  cpsr & ~tf_mask;                          /* Always in ARM mode */
      put_reg(14, get_reg(15, FALSE)-4, FALSE);
      put_reg(15, 8, FALSE);
      break;
    }
  }
}

/*----------------------------------------------------------------------------*/

void breakpoint()                              /* Looks like a prefetch abort */
{
spsr[abt_mode] = cpsr;
cpsr = (cpsr & ~mode_mask & ~tf_mask) | abt_mode;
put_reg(14, get_reg(15, FALSE)-4, FALSE);
put_reg(15, 12, FALSE);
return;
}

/*----------------------------------------------------------------------------*/

void undefined()
{
spsr[undef_mode] = cpsr;
cpsr = (cpsr & ~mode_mask & ~tf_mask) | undef_mode;
put_reg(14, get_reg(15, FALSE) - 4, FALSE);
put_reg(15, 4, FALSE);
return;
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void set_flags(int operation, int a, int b, int rd, int carry)
{
set_NZ(rd);
set_CF(a, rd, carry);
switch (operation)
  {
  case 1: set_VF_ADD(a, b, rd); break;
  case 2: set_VF_SUB(a, b, rd); break;
  default: printf("Flag setting error\n"); break;
  }
return;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void set_NZ(unsigned int value)
{
if (value == 0)            cpsr = cpsr | zf_mask; else cpsr = cpsr & ~zf_mask;
if ((value & bit_31) != 0) cpsr = cpsr | nf_mask; else cpsr = cpsr & ~nf_mask;
return;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void set_CF(unsigned int a, unsigned int rd, int carry)
{                                     /* Two ways result can equal an operand */
if ((rd > a) || ((rd == a) && (carry == 0))) cpsr = cpsr & ~cf_mask;
else                                         cpsr = cpsr |  cf_mask;
return;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void set_VF_ADD(int a, int b, int rd)
{
cpsr = (cpsr & ~vf_mask);                                         /* Clear VF */
if (((~(a ^ b) & (a ^ rd)) & bit_31) != 0) cpsr = cpsr | vf_mask;
return;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void set_VF_SUB(int a, int b, int rd)
{
cpsr = (cpsr & ~vf_mask);                                         /* Clear VF */
if ((((a ^ b) & (a ^ rd)) & bit_31) != 0) cpsr = cpsr | vf_mask;
return;
}

/*----------------------------------------------------------------------------*/

int check_cc(int condition)                   /*checks CC against flag status */
{
int go;

switch (condition & 0XF)
  {
  case 0X0: go = zf(cpsr);      break;
  case 0X1: go = not(zf(cpsr)); break;
  case 0X2: go = cf(cpsr);      break;
  case 0X3: go = not(cf(cpsr)); break;
  case 0X4: go = nf(cpsr);      break;
  case 0X5: go = not(nf(cpsr)); break;
  case 0X6: go = vf(cpsr);      break;
  case 0X7: go = not(vf(cpsr)); break;
  case 0X8: go = and(cf(cpsr),not(zf(cpsr))); break;
  case 0X9: go = or(not(cf(cpsr)),zf(cpsr));  break;
  case 0XA: go = not(xor(nf(cpsr),vf(cpsr))); break;
  case 0XB: go = xor(nf(cpsr),vf(cpsr));      break;
  case 0XC: go = and(not(zf(cpsr)),not(xor(nf(cpsr),vf(cpsr)))); break;
  case 0XD: go = or(zf(cpsr),xor(nf(cpsr),vf(cpsr)));            break;
  case 0XE: go = TRUE;  break;
  case 0XF: go = FALSE; break;
  }
return go;
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void state()
{
int i;

for (i=0; i<=7; i++) printf("%08X ", r[i]);
printf("\n");
for (i=8; i<=15; i++) printf("%08X ", r[i]);
printf("\n");

/* flags */
if ((cpsr & nf_mask) == 0) printf("."); else printf("N");
if ((cpsr & zf_mask) == 0) printf("."); else printf("Z");
if ((cpsr & cf_mask) == 0) printf("."); else printf("C");
if ((cpsr & vf_mask) == 0) printf("."); else printf("V");
printf("  ");

/* interrupts */
if ((cpsr & if_mask) == 0) printf("."); else printf("I");
if ((cpsr & ff_mask) == 0) printf("."); else printf("F");

/* architecture */
if ((cpsr & tf_mask) == 0) printf("  A"); else printf("  T");

/* mode */
/*printf("   Mode = ");*/
switch (cpsr & mode_mask)
  {
  case user_mode:  printf("  User");       break;
  case fiq_mode:   printf("  FIQ");        break;
  case irq_mode:   printf("  IRQ");        break;
  case sup_mode:   printf("  Supervisor"); break;
  case abt_mode:   printf("  Abort");      break;
  case undef_mode: printf("  Undefined");  break;
  default: printf("  Mode %02X", cpsr & mode_mask); break;
  }
printf("\n");

for (i=8; i<=14; i++) printf("%08X ", fiq_r[i-8]);
print_flags(spsr[0X11]);

printf("IRQ      ");
printf("         ");printf("         ");
printf("         ");printf("         ");
for (i=13; i<=14; i++) printf("%08X ", irq_r[i-13]);
print_flags(spsr[0X12]);

printf("Supervisor");
printf("        "); printf("         ");
printf("         ");printf("         ");
for (i=13; i<=14; i++) printf("%08X ", sup_r[i-13]);
print_flags(spsr[0X13]);

printf("Abort    ");
printf("         ");printf("         ");
printf("         ");printf("         ");
for (i=13; i<=14; i++) printf("%08X ", abt_r[i-13]);
print_flags(spsr[0X17]);

printf("Undefined");
printf("         ");printf("         ");
printf("         ");printf("         ");
for (i=13; i<=14; i++) printf("%08X ", undef_r[i-13]);
print_flags(spsr[0X1B]);

printf("\n");

return;
}

/*----------------------------------------------------------------------------*/

void print_flags(unsigned int flags)
{

if ((flags & nf_mask) == 0) printf("."); else printf("N");
if ((flags & zf_mask) == 0) printf("."); else printf("Z");
if ((flags & cf_mask) == 0) printf("."); else printf("C");
if ((flags & vf_mask) == 0) printf("."); else printf("V");
printf("  ");
if ((flags & if_mask) == 0) printf("."); else printf("I");
if ((flags & ff_mask) == 0) printf("."); else printf("F");
printf("\n");

return;
}

/*----------------------------------------------------------------------------*/

int not(int x)
{if (x==TRUE) return FALSE; else return TRUE;}

int and(int x, int y)
{if ((x==TRUE) && (y==TRUE)) return TRUE; else return FALSE;}

int or(int x, int y)
{if ((x==TRUE) || (y==TRUE)) return TRUE; else return FALSE;}

int xor(int x, int y)
{if (((x==TRUE) && (y==FALSE)) || ((x==FALSE) && (y==TRUE))) return TRUE;
else  return FALSE;}

/*----------------------------------------------------------------------------*/

int zf(int cpsr)
{if ((zf_mask & cpsr) != 0) return TRUE; else return FALSE;}

int cf(int cpsr)
{if ((cf_mask & cpsr) != 0) return TRUE; else return FALSE;}

int nf(int cpsr)
{if ((nf_mask & cpsr) != 0) return TRUE; else return FALSE;}

int vf(int cpsr)
{if ((vf_mask & cpsr) != 0) return TRUE; else return FALSE;}

/*----------------------------------------------------------------------------*/

int get_reg(int reg_no, int force_user)
{
int mode, value;

if (force_user == TRUE) mode = user_mode;
else                    mode = cpsr & mode_mask;

if (reg_no != 15)
  {
  switch (mode)
    {
    case user_mode:
      value = r[reg_no];
      break;

    case fiq_mode:
      if (reg_no < 8) value = r[reg_no];
      else            value = fiq_r[reg_no-8];
      break;

    case irq_mode:
      if (reg_no < 13) value = r[reg_no];
      else             value = irq_r[reg_no-13];
      break;

    case sup_mode:
      if (reg_no < 13) value = r[reg_no];
      else             value = sup_r[reg_no-13];
      break;

    case abt_mode:
      if (reg_no < 13) value = r[reg_no];
      else             value = abt_r[reg_no-13];
      break;

    case undef_mode:
      if (reg_no < 13) value = r[reg_no];
      else             value = undef_r[reg_no-13];
      break;
    }
  }
else                                                             /* PC access */
  value = r[15] + instruction_length();
         /* PC := PC+4 (or +2) at start of execution */

return value;
}

/*----------------------------------------------------------------------------*/

void put_reg(int reg_no, int value, int force_user)
{
int mode;

if (force_user == TRUE) mode = user_mode;
else                    mode = cpsr & mode_mask;

if (reg_no != 15)
  {
  switch (mode)
    {
    case user_mode:
      r[reg_no] = value;
      break;

    case fiq_mode:
      if (reg_no < 8) r[reg_no]   = value;
      else        fiq_r[reg_no-8] = value;
      break;

    case irq_mode:
      if (reg_no < 13) r[reg_no]    = value;
      else         irq_r[reg_no-13] = value;
      break;

    case sup_mode:
      if (reg_no < 13) r[reg_no]    = value;
      else         sup_r[reg_no-13] = value;
      break;

    case abt_mode:
      if (reg_no < 13) r[reg_no]    = value;
      else         abt_r[reg_no-13] = value;
      break;

    case undef_mode:
      if (reg_no < 13) r[reg_no]    = value;
      else       undef_r[reg_no-13] = value;
      break;
    }
  }
else
  r[15] = value & 0XFFFFFFFE;      /* Lose bottom bit, but NOT mode specific! */

return;
}

/*----------------------------------------------------------------------------*/

int instruction_length()
{if ((cpsr & tf_mask) == 0) return 4; else return 2;}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

unsigned int fetch()
{
unsigned int op_code;

op_code = read_mem((get_reg(15, FALSE) - instruction_length()),
                    instruction_length(), FALSE, FALSE, mem_instruction);
{
int i;
for (i=0;i<32;i++)
  if (past_opc_addr[i] == get_reg(15, FALSE)-instruction_length())
    {
    past_count++;
    i = 32;                                                  /* bodged escape */
    }

past_opc_addr[past_opc_ptr++] = get_reg(15, FALSE) - instruction_length();
past_opc_ptr = past_opc_ptr % past_size;
}

return op_code;
}

/*----------------------------------------------------------------------------*/

void inc_pc()
{
/* get_reg returns PC+4 for ARM & PC+2 for THUMB */

/*printf("get PC: %08x\n", get_reg(15, FALSE) ); */

put_reg(15, get_reg(15, FALSE), FALSE);

return;
}

/*----------------------------------------------------------------------------*/

void endian_swap(unsigned int start, unsigned int end)
{
unsigned int i, j;

for (i=start; i<end; i++)
  {
  j = getmem32(i);
  setmem32(i, ((j>>24) & 0X000000FF) |
              ((j>> 8) & 0X0000FF00) |
              ((j<< 8) & 0X00FF0000) |
              ((j<<24) & 0XFF000000));
  }
return;
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

int read_mem(unsigned int address, int size, int sign, int T, int source)
{
int data, alignment;

if ((address>>2) < mem_size)
  {
  alignment = address & 0X00000003;
  data      = getmem32(address>>2);

  switch (address & 0X00000003)                             /* Apply rotation */
    {
    case 0: break;                                                   /* RR  0 */
    case 1: data = (data << 24) | ((data >>  8) & 0X00FFFFFF); break;/* RR  8 */
    case 2: data = (data << 16) | ((data >> 16) & 0X0000FFFF); break;/* RR 16 */
    case 3: data = (data <<  8) | ((data >> 24) & 0X000000FF); break;/* RR 24 */
    }

  switch (size)
    {
    case 0: data = 0; break;                            /* A bit silly really */

    case 1:                                                    /* byte access */
      if ((sign==TRUE) && ((data & 0X00000080) != 0)) data = data | 0XFFFFFF00;
      else                                            data = data & 0X000000FF;
      break;

    case 2:                                               /* half-word access */
      if ((sign==TRUE) && ((data & 0X00008000) != 0)) data = data | 0XFFFF0000;
      else                                            data = data & 0X0000FFFF;
      break;

    case 4: break;                                             /* word access */

    default: printf("Illegally sized memor y read\n");
    }
  }
else
  {
  data = 0X12345678;
  printf("Reading %08X  data = %08X\n", address, data);
  print_out = TRUE;
  }

/*
if (T == TRUE) printf("User space forced\n");
*/

/*
switch(source)
{
case 0: break;
case 1: printf("Instruction address %08X data %08X\n", address, data); break;
case 2: printf("Data read   address %08X data %08X\n", address, data); break;
}
*/

return data;
}

/*----------------------------------------------------------------------------*/

void write_mem(unsigned int address, int data, int size, int T, int source)
{
unsigned int mask;

if ((address == tube_address) && (tube_address != 0))/* Deal with Tube output */
  {
  unsigned char c;

  c = data & 0XFF;

  if (print_out != TRUE)
    {
    if ((c == 0X0A) || (c == 0X0D)) printf("\n");
    else
      if ((c < 0X20) || (c >= 0X7F)) printf("%02X", c);
      else printf("%c", c);
    }
  else
    {
    printf("Tube output byte = %02X (", c);
    if ((c < 0X20) || (c >= 0X7F)) printf(".)\n", c); else printf("%c)\n", c);
    }
  }
else
  {
  if ((address>>2) < mem_size)
    {
    switch (size)
      {
      case 0: break;                                    /* A bit silly really */

      case 1:                                                  /* byte access */
        mask = 0X000000FF << (8*(address & 0X00000003));
        data = data << (8*(address & 0X00000003));
        setmem32(address >> 2, (getmem32(address >> 2) & ~mask) | (data & mask));
        break;

      case 2:                                            /* half-word acccess */
        mask = 0X0000FFFF << (8*(address & 0X00000002));
        data = data << (8*(address & 0X00000002));
         setmem32(address >> 2,(getmem32(address >> 2) & ~mask) | (data & mask));
        break;

      case 4:                                                  /* word access */
        setmem32(address>>2, data);
        break;

      default: printf("Illegally sized memor y write\n");
      }
    }
  else
    {
    printf("Writing %08X  data = %08X\n", address, data);
    print_out = TRUE;
    }

/*
if (T == TRUE) printf("User space forced\n");
*/

/*
switch(source)
{
case 0: break;
case 1: printf("Shouldn't happen\n"); break;
case 2: printf("Data write  address %08X data %08X\n", address, data); break;
}
*/
  }

return;
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void arm_disassemble(unsigned int address, unsigned int op_code)
{

unsigned int cc;

printf("%08X %08X ", address, op_code);

cc = (op_code>>28) & 0X0000000F;
switch ((op_code>>25) & 0X00000007)
  {
  case 0X0: dis_data_op(address, op_code, cc);break;
  case 0X1: dis_data_op(address, op_code, cc);break;
  case 0X2: dis_transfer(address, op_code, cc);break;
  case 0X3: dis_transfer(address, op_code, cc);break;
  case 0X4: dis_multiple(address, op_code, cc);break;
  case 0X5: dis_branch(address, op_code, cc);break;
  case 0X6: dis_coprocessor(address, op_code, cc);break;
  case 0X7: dis_system(address, op_code, cc);break;
  }

printf("\n");

return;
}

/*----------------------------------------------------------------------------*/

void dis_data_op(unsigned int address, unsigned int op_code, int cc)
{
int operation;
int shift_carry;

if (((op_code & mul_mask) == mul_op)                  /* Short multiplication */
 || ((op_code & long_mul_mask) == long_mul_op))   /*  and long multiplication */
  dis_mul(address, op_code, cc);
else
  if (is_it_sbhw(op_code) == TRUE) dis_sbhw(address, op_code, cc);
  else
    if ((op_code & swp_mask) == swp_op) dis_swp(address, op_code, cc);
    else
      {

      operation = (op_code & data_op_mask)>>21;

      if ((operation >= 0X8) && (operation <= 0XB) && ((op_code & s_mask) == 0))
        {                                              /* PSR transfers OR BX */
        if ((op_code & 0X0FBF0FFF) == 0X010F0000)                      /* MRS */
          dis_mrs(address, op_code, cc);
        else                                                           /* MSR */
          if (((op_code & 0X0DB6F000) == 0X0120F000)
           && ((op_code & 0X02000010) != 0X00000010))
            dis_msr(address, op_code, cc);
          else                                                      /* BX/BLX */
            if ((op_code & 0X012FFFD0) == 0X012FFF10)
              dis_bx(address, op_code, cc);
            else                                  /* Notice no condition code */
              if ((op_code & 0XFFF000F0) == 0XE1200070) printf("Breakpoint");
              else
                if ((op_code & 0X0FFF0FF0) == 0X016F0F10)              /* CLZ */
                  dis_dat(address, op_code, cc, operation);
                else printf("UNDEFINED");
        }
      else                                 /* data processing immediate shift */
        dis_dat(address, op_code, cc, operation);
      }

return;
}

/*----------------------------------------------------------------------------*/

void dis_mrs(unsigned int address, unsigned int op_code, int cc)
{
int spaces;

spaces = 2 + 1;
printf("MRS");

dis_cc(cc, &spaces);
dis_tab(spaces);

printf("R%d, ",(op_code & rd_mask)>>12);
if ((op_code & 0X00400000) == 0) printf("CPSR");
else                             printf("SPSR");

return;
}

/*----------------------------------------------------------------------------*/

void dis_msr(unsigned int address, unsigned int op_code, int cc)
{
int spaces;

spaces = 2 + 1;
printf("MSR");

dis_cc(cc, &spaces);
dis_tab(spaces);

if ((op_code & 0X00400000) == 0) printf("CPSR");
else                             printf("SPSR");

switch (op_code & 0X00090000)
  {
  case 0X00000000: printf("_none?"); break;
  case 0X00010000: printf("_mode");  break;
  case 0X00080000: printf("_flg");   break;
  case 0X00090000: break;
  }

printf(", ");

if ((op_code & imm_mask) == 0)                 /* Test applies for both cases */
  printf("R%d",(op_code & rm_mask));
else
  dis_immediate(op_code & 0X00000FFF);

return;
}

/*----------------------------------------------------------------------------*/

void dis_bx(unsigned int address, unsigned int op_code, int cc)
{
int spaces;

if ((op_code & 0X00000020) == 0) {printf("BX");  spaces = 4;}
else                             {printf("BLX"); spaces = 3;}

dis_cc(cc, &spaces);
dis_tab(spaces);

printf("R%d ", (op_code & rm_mask));
}

/*----------------------------------------------------------------------------*/

void dis_sbhw(unsigned int address, unsigned int op_code, int cc)
{
int spaces;

spaces = 2;

if ((op_code & 0X00100000) != 0) printf("LDR"); else printf("STR");

dis_cc(cc, &spaces);

switch (op_code & 0X00000060)
  {
  case 0X20: printf("H");  break;
  case 0X40: printf("SB"); break;
  case 0X60: printf("SH"); break;
  }

dis_tab(spaces);

printf("R%d, ", (op_code & rd_mask) >> 12);
printf("[R%d",  (op_code & rn_mask) >> 16);

if ((op_code & pre_mask) == 0)                                  /* Post index */
  {
  printf("]");
  dis_transfer_offset(op_code & 0X00000FFF, op_code & up_mask,
                      op_code & imm_hw_mask, TRUE);
  }
else                                                             /* Pre index */
  {
  dis_transfer_offset(op_code & 0X00000FFF, op_code & up_mask,
                      op_code & imm_hw_mask, TRUE);
  printf("]");
  if ((op_code & write_back_mask) != 0) printf("!");   /* Optional write back */
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_dat(unsigned int address, unsigned int op_code, int cc, int operation)
{
int spaces;

spaces = 2;                                                    /* Format info */

switch (operation)
  {
  case 0X0:printf("AND"); break;
  case 0X1:printf("EOR"); break;
  case 0X2:printf("SUB"); break;
  case 0X3:printf("RSB"); break;
  case 0X4:printf("ADD"); break;
  case 0X5:printf("ADC"); break;
  case 0X6:printf("SBC"); break;
  case 0X7:printf("RSC"); break;
  case 0X8:printf("TST"); break;
  case 0X9:printf("TEQ"); break;
  case 0XA:printf("CMP"); break;
  case 0XB:
    if ((op_code & s_mask) != 0) printf("CMN");
    else                         printf("CLZ");
    break;
  case 0XC:printf("ORR"); break;
  case 0XD:printf("MOV"); break;
  case 0XE:printf("BIC"); break;
  case 0XF:printf("MVN"); break;
  }

dis_cc(cc, &spaces);

switch (operation)
  {
  case 0X0: case 0X1: case 0X2: case 0X3:
  case 0X4: case 0X5: case 0X6: case 0X7:
  case 0XC: case 0XD: case 0XE: case 0XF:
    if ((op_code & s_mask) != 0) printf("S"); else spaces++;
    break;

  case 0X8: case 0X9: case 0XA: case 0XB:
    if ((op_code & rd_mask) == 0XF) printf("P"); else spaces++;
    break;
  }

dis_tab(spaces);

if ((op_code & 0X0FFF0FF0) == 0X016F0F10)
  {
  operation = 13;                              /* CLZ disassembles like a MOV */
  op_code = (op_code & 0XF000F00F) | 0X01A00000;           /*  with no shifts */
  };

switch (operation)
  {
  case 0X0: case 0X1: case 0X2: case 0X3:
  case 0X4: case 0X5: case 0X6: case 0X7:
  case 0XC:           case 0XE:
    printf("R%d, ", (op_code & rd_mask)>>12);
    printf("R%d, ", (op_code & rn_mask)>>16);
  break;

  case 0X8: case 0X9: case 0XA: case 0XB:
    printf("R%d, ", (op_code & rn_mask)>>16);
  break;

  case 0XD:                     case 0XF:
    printf("R%d, ", (op_code & rd_mask)>>12);
   break;

  }

if ((op_code & imm_mask) == 0) dis_shift(op_code);
else                           dis_immediate(op_code & 0X00000FFF);

return;
}

/*----------------------------------------------------------------------------*/

void dis_shift(unsigned int op_code)
{
int reg, type, distance, rs;

reg = op_code & 0X00F;                                     /* Register number */
type = (op_code & 0X060)>>5;                                 /* Type of shift */
if ((op_code & 0X010) == 0)
  {                                                        /* Immediate value */
  distance = (op_code & 0XF80)>>7;
  if (distance == 0)
    {
    if (type == 3) type = 4;                                           /* RRX */
    else if (type != 0) distance = 32;
    }

  switch (type)
    {
    case 0X0:
      if (distance == 0) printf("R%d", reg);
      else               printf("R%d LSL #%d", reg, distance);
    break;
    case 0X1:printf("R%d LSR #%d", reg, distance); break;
    case 0X2:printf("R%d ASR #%d", reg, distance); break;
    case 0X3:printf("R%d ROR #%d", reg, distance); break;
    case 0X4:printf("R%d RRX #1",  reg);           break;
    }
  }
else
  {
  rs = (op_code >> 8) & 0XF;
  switch (type)
    {
    case 0X0:printf("R%d LSL R%d", reg, rs); break;
    case 0X1:printf("R%d LSR R%d", reg, rs); break;
    case 0X2:printf("R%d ASR R%d", reg, rs); break;
    case 0X3:printf("R%d ROR R%d", reg, rs); break;
    }
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_immediate(unsigned int op2)
{
unsigned int dummy;

printf("#%X", ror(op2 & 0X0FF, (op2 & 0XF00)>>7, &dummy));
return;
}

/*----------------------------------------------------------------------------*/

void dis_mul(unsigned int address, unsigned int op_code, int cc)
{
int spaces;

if ((op_code & mul_long_bit) != 0)
  {
  spaces = 0;
  if ((op_code & mul_sign_bit) != 0) printf("S");
  else                               printf("U");

  if ((op_code & mul_acc_bit) != 0) printf("MLAL");
  else                              printf("MULL");

  dis_cc(cc, &spaces);

  if ((op_code & s_mask) != 0) printf("S"); else spaces++;

  dis_tab(spaces);

  printf("R%d, ", (op_code & rd_mask)>>12);
  printf("R%d, ", (op_code & rn_mask)>>16);
  printf("R%d, ", (op_code & rm_mask));
  printf("R%d",   (op_code & rs_mask)>>8);
  }
else
  {
  spaces = 2;
  if ((op_code & mul_acc_bit) != 0) printf("MLA");
  else                              printf("MUL");

  dis_cc(cc, &spaces);

  if ((op_code & s_mask) != 0) printf("S"); else spaces++;

  dis_tab(spaces);

  printf("R%d, ", (op_code & rn_mask)>>16);
  printf("R%d, ", (op_code & rm_mask));
  printf("R%d",   (op_code & rs_mask)>>8);
  if ((op_code & mul_acc_bit) != 0) printf(", R%d", (op_code & rd_mask)>>12);
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_swp(unsigned int address, unsigned int op_code, int cc)
{
int spaces;

spaces = 2;

printf("SWP");
dis_cc(cc, &spaces);

if ((op_code & byte_mask) != 0) printf("B");
else                            spaces++;
dis_tab(spaces);

printf("R%d, R%d, [R%d]", (op_code & rd_mask)>>12,
                          (op_code & rm_mask),
                          (op_code & rn_mask)>>16);
return;
}

/*----------------------------------------------------------------------------*/

void dis_transfer(unsigned int address, unsigned int op_code, int cc)
{
int spaces;

spaces = 2;

if ((op_code & undef_mask) == undef_code)
  printf("Undefined");
else
  {
  if ((op_code & load_mask) == 0) printf("STR");
  else                            printf("LDR");

  dis_cc(cc, &spaces);

  if ((op_code & byte_mask) == 0) spaces++; else printf("B");

  if (((op_code & pre_mask) == 0) && ((op_code & write_back_mask) != 0))
    {printf("T"); spaces--;}

  dis_tab(spaces);

  printf("R%d, ",(op_code & rd_mask)>>12);

/* Trap out PC relative as "absolute" */
  if ((op_code & 0X032F0000) == 0X010F0000)
    {
    int offset;
    offset = op_code & 0X00000FFF;
    if ((op_code & 0X00800000) == 0) offset = -offset;
    printf("%08X", get_reg(15, FALSE) + offset + 4); /* +4 for PC offset (ME) */
    }
  else
    {
    printf("[R%d",(op_code & rn_mask)>>16);
    if ((op_code & pre_mask) == 0)
      {
      printf("]");
      dis_transfer_offset(op_code & 0X00000FFF, op_code & up_mask,
                         ~op_code & imm_mask, FALSE); /* NB Imm. bit inverted */
      }
    else
      {
      dis_transfer_offset(op_code & 0X00000FFF, op_code & up_mask,
                         ~op_code & imm_mask, FALSE); /* NB Imm. bit inverted */
      printf("]");
      if ((op_code & write_back_mask) != 0) printf("!");
      }
    }
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_transfer_offset(int op2, int add, int imm, int sbhw)
{
/* sbhw bit 22 = 1 imm, = 0 reg
        others bit 25 = 0 imm, 1 = reg */

if ((imm == 0)                             /* Register offset so always print */
|| ((op2 & 0XF0F) != 0)                              /* (Any) offset non-zero */
|| (((op2 & 0X0F0) != 0) && (sbhw != TRUE)))
  {                         /* Check last nibble for `normal' addressing mode */
  printf(", ");
  if (add == 0) printf("-"); else printf("+");

  if (sbhw == 0)                                      /* word - unsigned byte */
    {
    if (imm == 0) dis_shift(op2);                                  /* bit(25) */
    else printf("#%X", (op2 & 0XFFF));
    }
  else
    {
    if (imm != 0)                                    /* half word signed byte */
      printf("#%02X", ((op2 & 0XF00) >> 4) | (op2 & 0X00F));
    else
      dis_shift(op2);
    }
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_multiple(unsigned int address, unsigned int op_code, int cc)
{
int spaces;

spaces = 2 - 1;

if ((op_code & load_mask) == 0) printf("STM");
else                            printf("LDM");

dis_cc(cc, &spaces);

switch ((op_code & 0X01800000)>>23)
  {
  case 0: printf("DA"); break;
  case 1: printf("IA"); break;
  case 2: printf("DB"); break;
  case 3: printf("IB"); break;
  }

dis_tab(spaces);

printf("R%d",(op_code & rn_mask)>>16);
if ((op_code & write_back_mask) != 0) printf("!");

printf(", {"); print_reg_list(op_code & 0X0000FFFF); printf("}");

if ((op_code & user_mask) != 0) printf("^");

return;
}

/*----------------------------------------------------------------------------*/

void dis_branch(unsigned int address, unsigned int op_code, int cc)
{
int offset;
int spaces;

offset = (op_code & branch_field) << 2;

if ((op_code & 0XF0000000) == 0XF0000000)                    /* Check for BLX */
  {
  printf("BLX     ");
  offset = offset | ((op_code >> 23) & 2);                    /* Get bit 1 in */
  }
else
  {
  spaces = 4;
  printf("B");
  if ((op_code & link_mask) != 0) printf("L"); else spaces++;

  dis_cc(cc, &spaces);
  dis_tab(spaces);
  }

if ((op_code & branch_sign) != 0)
  offset = offset | (~(branch_field<<2) & 0XFFFFFFFC);         /* sign extend */

printf("%08X", offset + address + 8);

return;
}

/*----------------------------------------------------------------------------*/

void dis_coprocessor(unsigned int address, unsigned int op_code, int cc)
{
printf("coprocessor\n");
}

/*----------------------------------------------------------------------------*/

void dis_system(unsigned int address, unsigned int op_code, int cc)
{
if ((op_code & 0X0F000000) != 0X0F000000)
  printf("Illegal");
else
  {
  int spaces;

  spaces = 2+1;
  printf("SWI");
  dis_cc(cc, &spaces);
  dis_tab(spaces);

  printf("%06X", (op_code & 0X00FFFFFF));

  }
return;
}

/*----------------------------------------------------------------------------*/

void dis_cc(int cc, int *spaces)
{
switch (cc)
  {
  case 0X0:printf("EQ");break;
  case 0X1:printf("NE");break;
  case 0X2:printf("CS");break;
  case 0X3:printf("CC");break;
  case 0X4:printf("MI");break;
  case 0X5:printf("PL");break;
  case 0X6:printf("VS");break;
  case 0X7:printf("VC");break;
  case 0X8:printf("HI");break;
  case 0X9:printf("LS");break;
  case 0XA:printf("GE");break;
  case 0XB:printf("LT");break;
  case 0XC:printf("GT");break;
  case 0XD:printf("LE");break;
  case 0XE:*spaces = *spaces + 2;break;
  case 0XF:printf("NV");break;
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_tab(int spaces)
{
while (spaces > 0) {printf(" "); spaces--;}
return;
}

/*----------------------------------------------------------------------------*/

void btb_insert(unsigned int addr, unsigned int instr)
{
int i, j;
int found;
int offset;
unsigned int target;

/* Search for entry */
found = FALSE;
for (i=0; (i<btb_size) && (found != TRUE); i++)
  if ((btb_cam[i] == addr) && (btb_valid[i] == TRUE))
    {
    found = TRUE;
    j = i;
    }

if (found != TRUE)
  {
/* Allocate unused entries first (or don't)
  found = FALSE;
  for (i=0; (i<btb_size) && (found != TRUE); i++)
  if (btb_valid[i] != TRUE)
    {
    found = TRUE;
    j = i;
    }
*/

  if (found != TRUE)
    j = btb_allocate(replacement_type, addr);


  offset = instr & branch_field;
  if ((offset & branch_sign) != 0)
    offset = offset | ~branch_field;              /* sign extend */
  target = addr + (offset<<2) + 8;

/*%%%%*/
branch_array4[addr>>2] = TRUE;
btc_insert_count++;
btc_time[j] = branch_count;    /* for LRU */

  if ((instr & 0X01000000) == 0X01000000)
    {                                          /* Link bit set */
    btb_cam[j]   = addr;
    btb_ram[j]   = target;
    btb_valid[j] = TRUE;
    }
  else
    {                                              /* No link */
    if ((instr & 0X00800000) != 0)
      {                                          /* Backwards */
      btb_cam[j]   = addr;
      btb_ram[j]   = target;
      btb_valid[j] = TRUE;
      }
    else
      {                                           /* Forwards */
      btb_cam[j]   = addr;
      btb_ram[j]   = target;
      btb_valid[j] = TRUE;
      }
    }
  }

return;
}

/*----------------------------------------------------------------------------*/

int btb_allocate(int type, unsigned int addr)
{
switch (type)
  {
  case cyclic_replace:                                              /* Cyclic */
    btb_index = (btb_index + 1) % btb_size;
    break;

  case random_replace:                                              /* Random */
    btb_index = rand() % btb_size;                      /* CHANGED was random */
    break;

  case rand21_replace:                                       /* Biased random */
    btb_index = rand() % ((3*btb_size)/2);               /*CHANGED was random */
    btb_index = btb_index % btb_size;                 /* halves with 1:2 bias */
    break;

  case LRU_replace:
    {
    int i;

    btb_index = 0;

    for (i=0; i<btb_size; i++)
      if (btc_time[i] < btc_time[btb_index]) btb_index = i;
    }
    break;

  case direct_mapped16:
    btb_index = addr >> 2;
    btb_index = btb_index & 0X0000001F;
    break;
  }

return btb_index;
}

/*----------------------------------------------------------------------------*/

void btc_flush()
{
int i;

for (i=0; i<btb_size; i++)
  {
  btb_valid[i] = FALSE;
  btc_time[i] = 0;
  }

return;
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/



void monitor_branch_init()
{                                                /* Initialise branch monitor */
int i;

branch_count        = 0;
branch_link_count   = 0;
branch_nolink_count = 0;
branch_taken_count  = 0;
branch_lt_count     = 0;
branch_nlt_count    = 0;

branch_pt     = 0;
branch_pnt    = 0;
branch_npt    = 0;
branch_npnt   = 0;
branch_l_pt   = 0;
branch_l_pnt  = 0;
branch_l_npt  = 0;
branch_l_npnt = 0;

btc_flush();
btb_index = 0;

return;
}

/*----------------------------------------------------------------------------*/

void monitor_branch(unsigned int instr)
{                                                        /* Branch monitoring */
int i;
unsigned int address;

branch_predicted = FALSE;

address = get_reg(15, FALSE)-4;

for (i=0; i<btb_size; i++)
  {
  if (btb_cam[i] == address)
    if (btb_valid[i] == TRUE)                             /* Branch predicted */
      {
      branch_predicted = TRUE;

btc_time[i] = branch_count;                                        /* for LRU */

{int dummy; glob1 = glob1 + bit_count(btb_cam[i]^btb_ram[i], &dummy); glob2++;}

      if (check_cc(instr>>28) == TRUE)
        {
        branch_pt++;                                                 /* Taken */
        if ((instr & 0X01000000) == 0X01000000) branch_l_pt++;   /* with link */
        }
      else
        {
        branch_pnt++;                                            /* Not taken */
        if ((instr & 0X01000000) == 0X01000000) branch_l_pnt++;  /* with link */
        }
      }
  }

if (branch_array4[address>>2] == TRUE)
  {
  if (check_cc(instr>>28) == TRUE) should_predict++;
  }
else
  {
  if (((instr & 0X0E000000) == 0X0A000000) && (check_cc(instr>>28) == FALSE))
    should_predict++;
  }


if ((instr & 0X0E000000) == 0X0A000000)
  {
  branch_count++;

/*%%%*/
  if (branch_array3[address>>2] > 0)
    branch_array2[address>>2] = branch_count - branch_array1[address>>2];
  branch_array1[address>>2] = branch_count;
  branch_array3[address>>2]++;


  if (branch_predicted != TRUE)
    {
    if (check_cc(instr>>28) == TRUE) branch_npt++;                   /* Taken */
    else                             branch_npnt++;              /* Not taken */
    }

  if (check_cc(instr>>28) == TRUE) branch_taken_count++;

  if ((instr & 0X01000000) == 0X01000000)
    {
    branch_link_count++;
    if (check_cc(instr>>28) == TRUE)
      {
      branch_lt_count++;
      btb_insert(address, instr);
      }
    }
  else
    {
    branch_nolink_count++;
    if (check_cc(instr>>28) == TRUE)
      {
      branch_nlt_count++;
      btb_insert(address, instr);
      }
    }
  }

return;
}

/*----------------------------------------------------------------------------*/

unsigned int predict(unsigned int address)
{
int i, j;
unsigned int prediction;

prediction = address + 4;                                          /* Default */

for (i=0; i<btb_size; i++)
  if ((btb_cam[i] == address) && (btb_valid[i] == TRUE))  /* Branch predicted */
    prediction = btb_ram[i];

return prediction;
}

/*----------------------------------------------------------------------------*/

/* Look at return instructions */
void monitor_return(unsigned int address, unsigned int instr)
{
int i, j;

if ((instr & 0X0FE0FFFF) == 0X01A0F00E)
  {
  return_count++;
  if ((instr & 0XF0000000) == 0XE0000000)
    {
    return_count_unc++;
    for (i=0, j=FALSE; (i<return_buff_size) & (j!=TRUE); i++)
      {
      if (return_CAM[i] == address)
        {
        j = TRUE;
        returns_found++;
        }
      }
    if (j != TRUE)
      {
      return_CAM[return_alloc_ptr] = address;
      return_alloc_ptr = (return_alloc_ptr + 1) % return_buff_size;
      }
    }
  else
    return_count_cond++;
  }

return;
}

/*----------------------------------------------------------------------------*/

void monitor_pc(unsigned int address, unsigned int last)
{

if (address != last + 4)
  {
  non_seq++;
  if (branch_predicted == FALSE)
    last = prefetch(last+4, last);
  }
else
  {
  if (branch_predicted == TRUE)
    last = prefetch(predict(last), last);
  }

monitor_pc_1(address, last);

return;
}

/*----------------------------------------------------------------------------*/

void monitor_pc_1(unsigned int address, unsigned int last)
{

if (address != last + 4)                     Non_sequential++;
else
  {
  if (((address ^ last) & BTC_CAM_top) != 0) Wrapped_block++;
  else                                       Bottom_only++;
  }

return;
}

/*----------------------------------------------------------------------------*/

unsigned int prefetch(unsigned int addr, unsigned int last)
{
int i;
int op_code;

wrong++;

for (i=1; i<=prefetch_depth; i++)
  {
  monitor_pc_1(addr, last);
  last = addr;
  addr = predict(addr);

/*
printf("Prefetch    address %08X\n", last);
*/

  op_code = read_mem(last, 4, FALSE, FALSE, mem_instruction);

  }

return last;
}

/*----------------------------------------------------------------------------*/

/* This is filthy :-( */
int get_number(char *ptr)
{
int a;

a = 0;
while (*ptr == ' ') ptr++;                        /* Strip any leading spaces */

while (((*ptr >= '0') && (*ptr <= '9')) || ((*ptr >= 'A') && (*ptr <= 'F')))
  {
  if ((*ptr >= '0') && (*ptr <= '9')) a = 16*a + *(ptr++) - '0';
  else                                a = 16*a + *(ptr++) - 'A' + 10;
  }

return a;
}

/*----------------------------------------------------------------------------*/
/* As the compiler can't manage it ...                                        */

int lsl(int value, int distance, int *cf)         /* cf is -internal- Boolean */
{
int result;

if (distance != 0)
  {
  if (distance < 32)
    {
    result = value << distance;
    *cf = (((value << (distance-1)) & bit_31) != 0);
   }
  else
    {
    result = 0X00000000;
    if (distance == 32) *cf = ((value & bit_0) != 0);
    else                *cf = (0 != 0);             /* internal "false" value */
    }
  }
else
  result = value;

return result;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

int lsr(unsigned int value, int distance, int *cf)/* cf is -internal- Boolean */
{
unsigned int result, mask;

if (distance != 0)
  {
  if (distance < 32)
    {
    if (distance != 0) mask = ~(0XFFFFFFFF << (32-distance));
    else               mask =   0XFFFFFFFF;
    result = (value >> distance) & mask;             /* Enforce logical shift */
    *cf = (((value >> (distance-1)) & bit_0) != 0);
    }
  else
    {                             /* Make a special case because C is so crap */
    result = 0X00000000;
    if (distance == 32) *cf = ((value & bit_31) != 0);
    else                *cf = (0 != 0);             /* internal "false" value */
    }
  }
else
  result = value;

return result;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

int asr(int value, int distance, int *cf)         /* cf is -internal- Boolean */
{
int result;

if (distance != 0)
  {
  if (distance < 32)
    {
    result = value >> distance;
    if (((value & bit_31) != 0) && (distance != 0))
      result = result | (0XFFFFFFFF << (32-distance));
                                  /* Sign extend - I don't trust the compiler */
    *cf = (((value >> (distance-1)) & bit_0) != 0);
    }
  else
    {                             /* Make a special case because C is so crap */
    *cf = ((value & bit_31) != 0);
    if ((value & bit_31) == 0) result = 0X00000000;
    else                       result = 0XFFFFFFFF;
    }
  }
else
  result = value;

return result;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

int ror(unsigned int value, int distance, int *cf)/* cf is -internal- Boolean */
{
int result;

if (distance != 0)
  {
  distance = distance & 0X1F;
  result = lsr(value, distance, cf)
         | lsl(value, 32 - distance, cf);            /* cf acts as dummy here */
  *cf = (((value >> (distance-1)) & bit_0) != 0);
  }
else
  result = value;

return result;
}

/*----------------------------------------------------------------------------*/
/*-------------------------THUMB disassemble ---------------------------------*/

void thumb_disassemble(unsigned int address, unsigned int op_code)
{
op_code = op_code & 0X0000FFFF;

printf("%08X     %04X ", address, op_code);
/*
        printchar(op_code & 0XFF);
        printchar((op_code>>8) & 0XFF);
        printf(" ");
*/
switch (op_code & 0XE000)
  {
  case 0X0000: dis_data0(address, op_code);         break;
  case 0X2000: dis_data1(address, op_code);         break;
  case 0X4000: dis_data_transfer(address, op_code); break;
  case 0X6000: dis_transfer0(address, op_code);     break;
  case 0X8000: dis_transfer1(address, op_code);     break;
  case 0XA000: dis_sp_pc(address, op_code);         break;
  case 0XC000: dis_lsm_b(address, op_code);         break;
  case 0XE000: dis_thumb_branch(address, op_code);  break;
  }
printf("\n");

return;
}

/*----------------------------------------------------------------------------*/

void case0_addr(unsigned int op_code, char *mnemonic)
{
int temp;

printf("%s     R%d, R%d, #", mnemonic, op_code & 7, (op_code >> 3) & 7);
temp = (op_code>>6) & 0X1F;
if (((op_code & 0X1800) != 0) && (temp == 0)) temp = 32;     /* Special cases */
printf("%02X", temp);
}

/*----------------------------------------------------------------------------*/

void dis_data0(unsigned int address, unsigned int op_code)
{
char mnemonic;

switch(op_code & 0X1800)
  {
  case 0X0000: case0_addr(op_code, "LSL"); break;
  case 0X0800: case0_addr(op_code, "LSR"); break;
  case 0X1000: case0_addr(op_code, "ASR"); break;
  case 0X1800:
    if ((op_code & 0X0200) == 0) printf("ADD"); else printf("SUB");
    printf("     R%d, R%d, ", op_code & 7, ((op_code >> 3) & 7));
    if ((op_code & 0X0400) == 0) printf("R"); else printf("#");
    printf("%d", ((op_code>>6) & 7));
    break;
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_data1(unsigned int address, unsigned int op_code)
{
switch(op_code & 0X1800)
  {
  case 0X0000: printf("MOV"); break;
  case 0X0800: printf("CMP"); break;
  case 0X1000: printf("ADD"); break;
  case 0X1800: printf("SUB"); break;
  }

printf("     R%d, #%02X", (op_code >> 8) & 7, op_code & 0XFF);

return;
}

/*----------------------------------------------------------------------------*/

void dis_data_transfer(unsigned int address, unsigned int op_code)
{
if ((op_code &0X1000) == 0)                                    /* load /store */
  {
  if ((op_code & 0X0800) == 0)
    {
    if ((op_code & 0X0400) == 0)                           /* data processing */
      {
      switch(op_code & 0X03C0)                      /* data processing opcode */
        {
        case 0X0000: printf("AND"); break;
        case 0X0040: printf("EOR"); break;
        case 0X0080: printf("LSL"); break;
        case 0X00C0: printf("LSR"); break;
        case 0X0100: printf("ASR"); break;
        case 0X0140: printf("ADC"); break;
        case 0X0180: printf("SBC"); break;
        case 0X01C0: printf("ROR"); break;
        case 0X0200: printf("TST"); break;
        case 0X0240: printf("NEG"); break;
        case 0X0280: printf("CMP"); break;
        case 0X02C0: printf("CMN"); break;
        case 0X0300: printf("ORR"); break;
        case 0X0340: printf("MUL"); break;
        case 0X0380: printf("BIC"); break;
        case 0X03C0: printf("MVN"); break;
        }
      printf("     R%d, R%d", op_code & 7, (op_code>>3) & 7);
      }
    else                                            /*special data processing */
      {
      switch(op_code & 0X0300)
        {
        case 0X0000:
          printf("ADD     R%d, R%d", ((op_code & 0X0080) >> 4) | (op_code & 7),
                                      (op_code >> 3) & 15);
          break;

        case 0X0100:
          printf("CMP     R%d, R%d", ((op_code & 0X0080) >> 4) | (op_code & 7),
                                      (op_code >> 3) & 15);
          break;

        case 0X0200:
          printf("MOV     R%d, R%d", ((op_code & 0X0080) >> 4) | (op_code & 7),
                                      (op_code >> 3) & 15);
          break;

        case 0X0300:                                                /* BX/BLX */
          if ((op_code & 0X0080) == 0) printf("BX      R%d", (op_code>>3) & 15);
          else                         printf("BLX     R%d", (op_code>>3) & 15);
          break ;
        }
      }                                     /* End of special data processing */
    }
  else                                              /* Load from literal pool */
    {
    printf("LDR     R%d, [PC, #%03X]", (op_code >> 8) & 7,
                                       (op_code & 0X00FF) << 2);
    printf("   (%08X)", (address & 0XFFFFFFFC) + 4 + ((op_code & 0x00FF) << 2));
    }
  }
else                           /* load/store word, halfword, byte, signed byte*/
  {
  switch(op_code & 0X0E00)
    {
    case 0X0000: printf("STR  "); break;
    case 0X0200: printf("STRH "); break;
    case 0X0400: printf("STRB "); break;
    case 0X0600: printf("LDRSB"); break;
    case 0X0800: printf("LDR  "); break;
    case 0X0A00: printf("LDRH "); break;
    case 0X0C00: printf("LDRB "); break;
    case 0X0E00: printf("LDRSH"); break;
    }

  printf("   R%d, [R%d, R%d]", (op_code & 7), ((op_code >> 3) & 7),
                                              ((op_code >> 6) & 7));
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_transfer0(unsigned int address, unsigned int op_code)
                                                  /* load/store w,b immediate */
{
if ((op_code & 0X0800) == 0) printf("STR"); else printf("LDR");
if ((op_code & 0X1000) == 0)
  printf("     R%d, [R%d, #%02X]", op_code & 7, (op_code >> 3) & 7,
                                                (op_code >> 4) & 0X07C);
else
  printf("B    R%d, [R%d, #%02X]", op_code & 7, (op_code >> 3) & 7,
                                                (op_code >> 6) & 0X1F);
return;
}

/*----------------------------------------------------------------------------*/

void dis_transfer1(unsigned int address, unsigned int op_code)
                                            /* load/store hw immediate & stack*/
{
if ((op_code & 0X0800) == 0) printf("STR"); else printf("LDR");

if ((op_code & 0X1000) == 0)
  printf("H    R%d, [R%d, #%02X]", op_code & 7, (op_code >> 3) & 7,
                                                (op_code >> 5) & 0X3E);
else
  printf("     R%d, [SP, #%03X]", (op_code >> 6) & 7, (op_code << 2) & 0X3FC);

return;
}

/*----------------------------------------------------------------------------*/

void dis_sp_pc(unsigned int address, unsigned int op_code)/* stack operations */
{
if ((op_code & 0X1000) == 0)                                  /* ADD SP or PC */
  {
  printf("ADD     R%d,", (op_code >> 8) & 7);

  if ((op_code & 0X0800) == 0) printf("PC"); else printf("SP");
  printf(", #%03X",  (op_code & 0X00FF) << 2);

  if ((op_code & 0X0800) == 0)
    printf("   (%08X)", (address & 0XFFFFFFFC) + 4 + ((op_code & 0X00FF) << 2));
  }
else
  {
  switch(op_code & 0X0F00)
    {
    case 0X0000:                                                 /* Adjust SP */
      if ((op_code & 0X0080) == 0) printf("ADD");
      else                         printf("SUB");
      printf("     SP, SP, #%03X", (op_code & 0X7F) << 2);
      break;

    case 0X0400:
    case 0X0500:
    case 0X0C00:
    case 0X0D00:
      if ((op_code & 0X0800) == 0) printf("PUSH");
      else                         printf("POP ");
      printf("    {");
      print_reg_list(op_code & 0X00FF);

      if ((op_code & 0X0100) != 0)
        {
        if ((op_code & 0X00FF) != 0) printf(",");
        if ((op_code & 0X0800) == 0) printf("LR");
        else                         printf("PC");
        }
      printf("}");
      break;

    case 0X0E00:
      printf("Breakpoint");
      break;

    case 0X0100:
    case 0X0200:
    case 0X0300:
    case 0X0600:
    case 0X0700:
    case 0X0800:
    case 0X0900:
    case 0X0A00:
    case 0X0B00:
    case 0X0F00:
      printf("Undefined");
      break;
    }
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_lsm_b(unsigned int address, unsigned int op_code)
{
if ((op_code & 0X1000) == 0)
  {
  if ((op_code & 0X0800) == 0) printf ("STM"); else printf("LDM");
  printf("IA   R%d", (op_code >> 8) & 7);

                                           /*option for writeback suppression */
  if (((op_code & 0X0800) == 0)
    || ((1 << ((op_code >> 8) & 7) & op_code) == 0)) printf("!");

  printf(", {"); print_reg_list(op_code & 0X00FF); printf("}");
  }
else
  {
  if ((op_code & 0X0F00) != 0X0F00)
    {
    int offset;

    switch (op_code & 0X0F00)
      {
      case 0X0000: printf("BEQ     "); break;
      case 0X0100: printf("BNE     "); break;
      case 0X0200: printf("BCS     "); break;
      case 0X0300: printf("BCC     "); break;
      case 0X0400: printf("BMI     "); break;
      case 0X0500: printf("BPL     "); break;
      case 0X0600: printf("BVS     "); break;
      case 0X0700: printf("BVC     "); break;
      case 0X0800: printf("BHI     "); break;
      case 0X0900: printf("BLS     "); break;
      case 0X0A00: printf("BGE     "); break;
      case 0X0B00: printf("BLT     "); break;
      case 0X0C00: printf("BGT     "); break;
      case 0X0D00: printf("BLE     "); break;
      case 0X0E00: printf("B       "); break;
      case 0X0F00: printf("wrong !");  break;
      }

    offset = (op_code & 0X00FF) << 1;
    if ((op_code & 0X0080) != 0) offset = offset | 0XFFFFFE00;
    printf("%08X", address + 4 + offset);
    }
  else printf("SWI     #%02X", op_code & 0XFF);
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_thumb_branch1(unsigned int op_code, int exchange)
{
int offset;

offset = BL_prefix << 12;
if ((BL_prefix & 0X0400) != 0) offset = offset | 0XFF800000;
offset = offset | ((op_code & 0X07FF) << 1);
if (exchange != TRUE) printf("BL      %08X", BL_address + offset);
else                  printf("BLX     %08X", (BL_address + offset) & 0XFFFFFFFC);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void dis_thumb_branch(unsigned int address, unsigned int op_code)
{
switch (op_code & 0X1800)
  {
  case 0X0000: printf("B       ");
    {
    int offset;
    offset = (op_code & 0X07FF) << 1;
    if ((op_code & 0X0400) != 0) offset = offset | 0XFFFFF000;
    printf("%08X", address + 4 + offset);
    }
    break;

  case 0X0800:
    if ((op_code & 0X0001) == 0X0000) dis_thumb_branch1(op_code, TRUE);
    else printf("Undefined");
    break;

  case 0X1000: printf("BL prefix");
    BL_prefix  = op_code & 0X07FF;
    BL_address = address + 4;
    break;

  case 0X1800:
    dis_thumb_branch1(op_code, FALSE);
    break;
  }

return;
}

/*----------------------------------------------------------------------------*/

void print_reg_list(int i)                      /* Assumes "i" already masked */
{
int state, k;

state = 0;                                            /* little state machine */
k     = 0;

while (i != 0)
  {
  switch (state)
    {
    case 0:
        if ((i & 1) != 0) {printf("R%d", k); state = 1;}
        break;
    case 1:
        if ((i & 1) != 0) {printf("-"); state = 2;}
        else              {printf(","); state = 0;}
        break;
    case 2:
        if ((i & 1) == 0)
          {
          printf("R%d", k - 1); if (i != 0) printf(",");
          state = 0;
          }
        break;
    }
  i = i >> 1;
  k = k + 1;
  }                                                           /* End of WHILE */

if (state==2) printf("R%d", k - 1);

return;
}

/*----------------------------------------------------------------------------*/

void printchar(char c)
{
if ((c < ' ') || (c > '~')) c = '.';
printf("%c, c");
return;
}

/*----------------------------------------------------------------------------*/

/*------------------------------ THUMB EXECUTE -------------------------------*/

/*----------------------------------------------------------------------------*/

void data0(unsigned int op_code)
{
unsigned int op2, rn;
unsigned int shift, result, cf;

rn = get_reg(((op_code >> 3) & 7), FALSE);           /* Called "Rm" in shifts */
shift = ((op_code >> 6) & 0X0000001F);             /* Extracted speculatively */

if ((op_code & 0X1800) != 0X1800)                                   /* Shifts */
  {
  cf = ((cpsr & cf_mask) != 0);                                    /* default */
  switch(op_code & 0X1800)
    {
    case 0X0000: result = lsl(rn, shift, &cf); break;              /* LSL (1) */
    case 0X0800:                                                   /* LSR (1) */
      if (shift == 0) shift = 32;
      result = lsr(rn, shift, &cf);
      break;
    case 0X1000:                                                   /* ASR (1) */
      if (shift == 0) shift = 32;
      result = asr(rn, shift, &cf);
      break;
    default: printf("This compiler is broken\n"); break;
    }

  if (cf) cpsr = cpsr |  cf_mask;
  else    cpsr = cpsr & ~cf_mask;
  set_NZ(result);
  put_reg((op_code & 7), result, FALSE);

  f1_a2count++;
  }
else
  {
  if ((op_code & 0X0400) == 0)                               /* ADD(3)/SUB(3) */
    {
    op2 = get_reg((op_code >> 6) & 7, FALSE);
    f2_a2count++;
    }
  else                                                       /* ADD(1)/SUB(1) */
    {
    op2 = (op_code >> 6) & 7;
    f2_a1count++;
    };

  if ((op_code & 0X0200) == 0)
    {result = rn + op2; set_flags(flag_add, rn, op2, result, 0);}
  else
    {result = rn - op2; set_flags(flag_sub, rn, op2, result, 1);}

  put_reg(op_code & 7, result, FALSE);
  }

return;
}

/*----------------------------------------------------------------------------*/

void data1(unsigned int op_code)
{
int rd, imm;
int result;

rd = (op_code >> 8) & 7;
imm = op_code & 0X00FF;

switch(op_code & 0X1800)
  {
  case 0X0000:                                                     /* MOV (1) */
    result = imm;
    set_NZ(result);
    put_reg(rd, result, FALSE);
    break;

  case 0X0800:                                                     /* CMP (1) */
    result = (get_reg(rd, FALSE) - imm);
    set_flags(flag_sub, get_reg(rd, FALSE), imm, result, 1);
    break;

  case 0X1000:                                                     /* ADD (2) */
    result = (get_reg(rd, FALSE) + imm);
    set_flags(flag_add, get_reg(rd, FALSE), imm, result, 0);
    put_reg(rd, result, FALSE);
    break;

  case 0X1800:                                                     /* SUB (2) */
    result = (get_reg(rd, FALSE) - imm);
    set_flags(flag_sub, get_reg(rd, FALSE), imm, result, 1);
    put_reg(rd, result, FALSE);
    break;
  }

f3_a1count++;

return;
}

/*----------------------------------------------------------------------------*/

void data_transfer(unsigned int op_code)
{
unsigned int rd, rm, mode;
int c_flag;
int cf;

unsigned int address;

signed int result;

if ((op_code &0X1000) == 0)                                 /* NOT load/store */
  {
  if ((op_code & 0X0800) == 0)                       /* NOT load literal pool */
    {
    if ((op_code & 0X0400) == 0)                           /* data processing */
      {
      rd = get_reg((op_code & 7), FALSE);
      rm = get_reg(((op_code >> 3) & 7), FALSE);

      switch(op_code & 0X03C0)                      /* data processing opcode */
        {
        case 0X0000:                                                   /* AND */
          result = rd & rm;
          put_reg(op_code & 7, result, FALSE);
          set_NZ(result);

          f4_a4count++;
          break;

        case 0X0040:                                                   /* EOR */
          result = rd ^ rm;
          put_reg(op_code & 7, result, FALSE);
          set_NZ(result);

          f4_a4count++;
          break;

        case 0X0080:                                               /* LSL (2) */
          cf = ((cpsr & cf_mask) != 0);                            /* default */
          result = lsl(rd, rm & 0X000000FF, &cf);
          if (cf) cpsr = cpsr |  cf_mask;
          else    cpsr = cpsr & ~cf_mask;
          set_NZ(result);
          put_reg(op_code & 7, result, FALSE);

          f4_a3count++;
          break;

        case 0X00C0:                                               /* LSR (2) */
          cf = ((cpsr & cf_mask) != 0);                            /* default */
          result = lsr(rd, rm & 0X000000FF, &cf);
          if (cf) cpsr = cpsr |  cf_mask;
          else    cpsr = cpsr & ~cf_mask;
          set_NZ(result);
          put_reg(op_code & 7, result, FALSE);

          f4_a3count++;
          break;

        case 0X0100:                                               /* ASR (2) */
          cf = ((cpsr & cf_mask) != 0);                            /* default */
          result = asr(rd, rm & 0X000000FF, &cf);
          if (cf) cpsr = cpsr |  cf_mask;
          else    cpsr = cpsr & ~cf_mask;
          set_NZ(result);
          put_reg(op_code & 7, result, FALSE);

          f4_a3count++;
          break;

        case 0X0140:                                                   /* ADC */
          result = rd + rm;
          if ((cpsr & cf_mask) != 0) result = result + 1;           /* Add CF */
          set_flags(flag_add, rd, rm, result, cpsr & cf_mask);
          put_reg(op_code & 7, result, FALSE);

          f4_a3count++;
          break;

        case 0X0180:                                                   /* SBC */
          result = rd - rm - 1;
          if ((cpsr & cf_mask) != 0) result = result + 1;
          set_flags(flag_sub, rd, rm, result, cpsr & cf_mask);
          put_reg(op_code & 7, result, FALSE);

          f4_a3count++;
          break;

        case 0X01C0:                                                   /* ROR */
          cf = ((cpsr & cf_mask) != 0);                            /* default */
          result = ror(rd, rm & 0X000000FF, &cf);
          if (cf) cpsr = cpsr |  cf_mask;
          else    cpsr = cpsr & ~cf_mask;
          set_NZ(result);
          put_reg(op_code & 7, result, FALSE);

          f4_a3count++;
          break;

        case 0X0200:                                                   /* TST */
          set_NZ(rd & rm);

          f4_a1count++;
          break;

        case 0X0240:                                                   /* NEG */
          result = -rm;
          put_reg(op_code & 7, result, FALSE);
          set_flags(flag_sub, 0, rm, result, 1);

          f4_a1count++;
          break;

        case 0X0280:                                               /* CMP (2) */
          set_flags(flag_sub, rd, rm, rd-rm, 1);

          f4_a1count++;
          break;

        case 0X02C0:                                                   /* CMN */
          set_flags(flag_add, rd, rm, rd+rm, 0);

          f4_a2count++;
          break;

        case 0X0300:                                                   /* ORR */
          result = rd | rm;
          set_NZ(result);
          put_reg(op_code & 7, result, FALSE);

          f4_a2count++;
          break;

        case 0X00340:                                                  /* MUL */
          result = rm * rd;
          set_NZ(result);
          put_reg(op_code & 7, result, FALSE);

          f4_a4count++;
          break;

        case 0X0380:                                                   /* BIC */
          result = rd & ~rm;
          set_NZ(result);
          put_reg(op_code & 7, result, FALSE);

          f4_a2count++;
          break;

        case 0X03C0:                                                   /* MVN */
          result = ~rm;
          set_NZ(result);
          put_reg(op_code & 7, result, FALSE);

          f4_a2count++;
          break;
        }                                                    /* End of switch */
      }
    else                                            /*special data processing */
      {                                                     /* NO FLAG UPDATE */
      switch(op_code & 0X0300)
        {
        case 0X0000:                                /* ADD (4) high registers */
          rd = ((op_code & 0X0080) >> 4) | (op_code & 7);
          rm = get_reg(((op_code >> 3) & 15), FALSE);
          put_reg(rd, get_reg(rd, FALSE) + rm, FALSE);

          f5_a2count++;
          break;

        case 0X0100:                                /* CMP (3) high registers */
          rd = get_reg((((op_code & 0X0080) >> 4) | (op_code & 7)), FALSE);
          rm = get_reg(((op_code >> 3) & 15), FALSE);
          set_flags(flag_sub, rd, rm, rd-rm, 1);
          f5_a2count++;
          break;

        case 0X0200:                                /* MOV (2) high registers */
          rd = ((op_code & 0X0080) >> 4) | (op_code & 7);
          rm = get_reg(((op_code >> 3) & 15), FALSE);

          if (rd == 15) rm = rm & 0XFFFFFFFE;              /* Tweak mov to PC */
          put_reg(rd, rm, FALSE);

          f5_a2count++;
          break;

        case 0X0300:                                             /* BX/BLX Rm */
          bx((op_code>>3) & 0XF, op_code & 0X0080);

          f5_a4count++;
          break;
        }                                                    /* End of switch */
      }
    }
  else                                              /* load from literal pool */
    {                                                               /* LDR PC */
    rd = ((op_code >> 8) & 7);
    address = (((op_code & 0X00FF) << 2)) + (get_reg(15, FALSE) & 0XFFFFFFFC);
    put_reg(rd, read_mem(address, 4, FALSE, FALSE, mem_data), FALSE);

    f6_a10count++;
    }
  }
else                          /* load/store word, halfword, byte, signed byte */
  {
  int rm, rn;
  int data;

  rd = (op_code & 7);
  rn = get_reg(((op_code >> 3) & 7), FALSE);
  rm = get_reg(((op_code >> 6) & 7), FALSE);

  switch(op_code & 0X0E00)
    {
    case 0X0000:                                          /* STR (2) register */
      write_mem(rn + rm, get_reg(rd, FALSE), 4, FALSE, mem_system);

      f7_a11count++;
      break;

    case 0X0200:                                         /* STRH (2) register */
      write_mem(rn + rm, get_reg(rd, FALSE), 2, FALSE, mem_system);

      f8_a13count++;
      break;

    case 0X0400:                                         /* STRB (2) register */
      write_mem(rn + rm, get_reg(rd, FALSE), 1, FALSE, mem_system);

      f8_a13count++;
      break;

    case 0X0600:                                            /* LDRSB register */
      data = read_mem(rn + rm, 1, TRUE, FALSE, mem_system);   /* Sign extends */
      put_reg(rd, data, FALSE);

      f8_a13count++;
      break;

    case 0X0800:                                          /* LDR (2) register */
      data = read_mem(rn + rm, 4, FALSE, FALSE, mem_system);
      put_reg(rd, data, FALSE);

      f7_a11count++;
      break;

    case 0X0A00:                                         /* LDRH (2) register */
      data = read_mem(rn + rm, 2, FALSE, FALSE, mem_system);  /* Zero extends */
      put_reg(rd, data, FALSE);

      f8_a13count++;
      break;

    case 0X0C00:                                                  /* LDRB (2) */
      data = read_mem(rn + rm, 1, FALSE, FALSE, mem_system);  /* zero extends */
      put_reg(rd, data, FALSE);

      f7_a11count++;
      break;

    case 0X0E00:                                                 /* LDRSH (2) */
      data = read_mem(rn + rm, 2, TRUE, FALSE, mem_system);    /* sign extend */
      put_reg(rd, data, FALSE);

      f8_a13count++;
      break;
    }
  }
return;
}

/*----------------------------------------------------------------------------*/

void transfer0(unsigned int op_code)
{
int rd, rn;
int location, data;

rn = get_reg(((op_code >> 3) & 7), FALSE);

if ((op_code & 0X0800) == 0)                                           /* STR */
  {
  rd = get_reg((op_code & 7), FALSE);
  if ((op_code & 0X1000) == 0)                           /* STR (1) 5-bit imm */
    {
    location = rn + ((op_code >> 4) & 0X07C);             /* shift twice = *4 */
    write_mem(location, rd, 4, FALSE, mem_system);
    }
  else                                                            /* STRB (1) */
    {
    location = rn + ((op_code >> 6) & 0X1F);
    write_mem(location, rd, 1, FALSE, mem_system);
    }
  }
else                                                               /* LDR (1) */
  {
  rd = op_code & 7;
  if ((op_code & 0X1000) == 0)
    {
    location = (rn + ((op_code >> 4) & 0X07C));           /* shift twice = *4 */
    data = read_mem(location, 4, FALSE, FALSE, mem_system);
    }
  else                                                            /* LDRB (1) */
    {
    location = (rn + ((op_code >> 6) & 0X1F));
    data = read_mem(location, 1, FALSE, FALSE, mem_system);  /* zero extended */
    }
  put_reg(rd, data, FALSE);
  }

f9_a10count++;

return;
}

/*----------------------------------------------------------------------------*/

void transfer1(unsigned int op_code)
{
int rd, rn;
int data, location;

switch(op_code & 0X1800)
  {
  case 0X0000:                                                    /* STRH (1) */
    rn = get_reg((op_code >> 3) & 7, FALSE);
    rd = op_code & 7;
    data = get_reg(rd, FALSE);
    location = rn + ((op_code >> 5) & 0X3E);                   /* x2 in shift */
    write_mem(location, data, 2, FALSE, mem_system);

    f10_a12count++;
    break;

  case 0X0800:                                                    /* LDRH (1) */
    rd = op_code & 7;
    rn = get_reg((op_code >> 3) & 7, FALSE);
    location = rn + ((op_code >> 5) & 0X3E);                   /* x2 in shift */
    data = read_mem(location, 2, FALSE, FALSE, mem_system);  /* zero extended */
    put_reg(rd, data, FALSE);

    f10_a12count++;
    break;

  case 0X1000:                                                 /* STR (3) -SP */
    data = get_reg(((op_code >> 8) & 7), FALSE);
    rn = get_reg(13, FALSE);                                            /* SP */
    location = rn + ((op_code & 0X00FF) * 4);
    write_mem(location, data, 4, FALSE, mem_system);

    f11_a10count++;
    break;

  case 0X1800:                                                  /* LDR (4) -SP*/
    rd = (op_code >> 8) & 7;
    rn = get_reg(13, FALSE);                                            /* SP */
    location = rn + ((op_code & 0X00FF) *4 );                  /* x2 in shift */
    data = read_mem(location, 4, FALSE, FALSE, mem_system);
    put_reg(rd, data, FALSE);

    f11_a10count++;
    break;

  }
return;
}

/*----------------------------------------------------------------------------*/

void sp_pc(unsigned int op_code)
{
int rd, sp, data;
int reg, reg_list, count, first_reg, address, base;

if ((op_code & 0X1000) == 0)                                  /* ADD SP or PC */
  {
  rd = (op_code >> 8) & 7;

  if ((op_code & 0X0800) == 0)                                  /* ADD(5) -PC */
    data = ((get_reg(15, FALSE) & 0XFFFFFFFC) + ((op_code & 0X00FF) << 2));
                                                    /* get_reg supplies PC +2 */
  else                                                          /* ADD(6) -SP */
    data = ((get_reg(13, FALSE)) + ((op_code & 0X00FF) << 2));

  put_reg(rd, data, FALSE);

  f12_a1count++;
  }
else                                                             /* Adjust SP */
  {
  switch(op_code & 0X0F00)
    {
    case 0X0000:
      if ((op_code & 0X0080) == 0)                              /* ADD(7) -SP */
        sp = get_reg(13, FALSE) + ((op_code & 0X7F) << 2);
      else                                                      /* SUB(4) -SP */
        sp = get_reg(13, FALSE) - ((op_code & 0X7F) << 2);

      put_reg(13, sp, FALSE);

      f13_a1count++;
      break;

    case 0X0400:
    case 0X0500:
    case 0X0C00:
    case 0X0D00:
      {
      int reg_list;

      reg_list = op_code & 0X000000FF;

      if ((op_code & 0X0800) == 0)                                    /* PUSH */
        {
        if ((op_code & 0X0100) != 0) reg_list = reg_list | 0X4000;
        stm(2, 13, reg_list, 1, 0);
        }
      else                                                             /* POP */
        {
        if ((op_code & 0X0100) != 0) reg_list = reg_list | 0X8000;
        ldm(1, 13, reg_list, 1, 0);
        }

      f14_a15count++;
      }
      break;

    case 0X0E00:                                                /* Breakpoint */
printf("Breakpoint\n");
      breakpoint();
      break;

    case 0X0100:
    case 0X0200:
    case 0X0300:
    case 0X0600:
    case 0X0700:
    case 0X0800:
    case 0X0900:
    case 0X0A00:
    case 0X0B00:
    case 0X0F00:
printf("Undefined\n");
      undefined();
      break;
  }

return;
}
}

/*----------------------------------------------------------------------------*/

void lsm_b(unsigned int op_code)
{
unsigned int offset;

if ((op_code & 0X1000) == 0)
  {
  if ((op_code & 0X0800) == 0)                                    /* STM (IA) */
    stm(1, (op_code >> 8) & 7, op_code & 0X000000FF, 1, 0);
  else                                                            /* LDM (IA) */
    ldm(1, (op_code >> 8) & 7, op_code & 0X000000FF, 1, 0);

  f15_a15count++;
  }
else                                               /* conditional BRANCH B(1) */
  {
  if ((op_code & 0X0F00) != 0X0F00)                      /* Branch, not a SWI */
    {
    if (check_cc(op_code>>8) == TRUE)
      {
      offset = (op_code & 0X00FF) << 1;
                                                               /* sign extend */
      if ((op_code & 0X0080) != 0) offset = offset | 0XFFFFFE00;

      put_reg(15, get_reg(15, FALSE) + offset, FALSE);
                                                   /* get_reg supplies pc + 2 */
      f16_a19count++;
/*
printf("%08X", address + 4 + offset);
*/
      }
    }
  else                                                                 /* SWI */
    {
    /* bodge op_code to pass only SWI No. N.B. no copro in Thumb */
    offset = op_code & 0X00FF;

    my_system(offset);
    }
  }
}


/*----------------------------------------------------------------------------*/

void thumb_branch1(unsigned int op_code, int exchange)
{
int offset, lr;

lr = get_reg(14, FALSE);                     /* Retrieve first part of offset */
offset = lr + ((op_code & 0X07FF) << 1);

lr = get_reg(15, FALSE) - 2 + 1;                /* + 1 to indicate Thumb mode */

if (exchange == TRUE)
  {
  cpsr = cpsr & ~tf_mask;                               /* Change to ARM mode */
  offset = offset & 0XFFFFFFFC;
  }

put_reg(15, offset, FALSE);
put_reg(14, lr, FALSE);

return;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void thumb_branch(unsigned int op_code)
{
int offset, lr;

switch (op_code & 0X1800)
  {
  case 0X0000:                                            /* B -uncond. B(2)  */
    offset = (op_code & 0X07FF) << 1;
    if ((op_code & 0X0400) != 0) offset = offset | 0XFFFFF000; /* sign extend */
    put_reg(15, (get_reg(15, FALSE) + offset), FALSE);

    f18_a19count++;
    break;

  case 0X0800:                                                         /* BLX */
    if ((op_code & 0X0001) == 0) thumb_branch1(op_code, TRUE);
    else printf("Undefined\n");
    break;

  case 0X1000:                                                   /* BL prefix */
    BL_prefix = op_code & 0X07FF;
    offset = BL_prefix << 12;

    if ((BL_prefix & 0X0400) != 0) offset = offset | 0XFF800000; /* Sign ext. */
    offset = get_reg(15, FALSE) + offset;
    put_reg(14, offset, FALSE);

    f19_a19count++;
    break;

  case 0X1800:                                                          /* BL */
    thumb_branch1(op_code, FALSE);
    break;
  }

return;
}

/*----------------------------------------------------------------------------*/

/*------------------------------ Charlies functions---------------------------*/

/*----------------------------------------------------------------------------*/
unsigned int getmem32(int number)
{
 number=number%RAMSIZE;
 return memory[(number<<2)] |
        memory[(number<<2)+1]<<8 |
        memory[(number<<2)+2]<<16 |
        memory[(number<<2)+3]<<24;
}

void setmem32(int number,unsigned int reg)
{
 number=number&(RAMSIZE-1);
 memory[(number<<2)+0] = (reg>>0)&0xff;
 memory[(number<<2)+1] = (reg>>8)&0xff;
 memory[(number<<2)+2] = (reg>>16)&0xff;
 memory[(number<<2)+3] = (reg>>24)&0xff;
}
